"""
This module handles the four types of loads in DStability.
"""
import abc
from typing import List, Optional

from pydantic import NoneStr, confloat, validator

from geolib.models import BaseDataClass

from ...geometry.one import Point
from .internal import (
    PersistableConsolidation,
    PersistableLineLoad,
    PersistablePoint,
    PersistableUniformLoad,
)


class DStabilityLoad(BaseDataClass):
    """Base Class for Loads."""

    label: NoneStr

    @abc.abstractmethod
    def to_internal_datastructure(self):
        raise NotImplementedError


class Consolidation(BaseDataClass):
    degree: confloat(ge=0, le=100) = 100
    layer_id: int

    def to_internal_datastructure(self) -> PersistableConsolidation:
        return PersistableConsolidation(Degree=self.degree, LayerId=str(self.layer_id))


class UniformLoad(DStabilityLoad):
    """UniformLoad."""

    start: float
    end: float
    magnitude: confloat(ge=0)
    angle_of_distribution: confloat(ge=0, le=90)

    @validator("end")
    def end_greater_than_start(cls, v, values):
        if v <= values["start"]:
            raise ValueError(f"End {v} should be greater than start ({values['start']})")
        return v

    def to_internal_datastructure(self) -> PersistableUniformLoad:
        return PersistableUniformLoad(
            Label=self.label,
            Start=self.start,
            End=self.end,
            Magnitude=self.magnitude,
            Spread=self.angle_of_distribution,
        )


class LineLoad(DStabilityLoad):
    """DStability Lineload."""

    location: Point
    angle: confloat(ge=-360, le=360)
    magnitude: confloat(ge=0)
    angle_of_distribution: confloat(ge=0, le=90)

    def to_internal_datastructure(self) -> PersistableLineLoad:
        return PersistableLineLoad(
            Label=self.label,
            Location=PersistablePoint(X=self.location.x, Z=self.location.z),
            Angle=self.angle,
            Magnitude=self.magnitude,
            Spread=self.angle_of_distribution,
        )


class TreeLoad(DStabilityLoad):
    """Inherits :class:`~geolib.models.dstability.loads.DStabilityLoad`."""

    height: float
    wind_force: float
    width_of_root_zone: float
    angle_of_distribution: float


class Earthquake(DStabilityLoad):
    """Inherits :class:`~geolib.models.dstability.loads.DStabilityLoad`."""

    horizontal_factor: float
    vertical_factor: float
    free_water_factor: float



import logging
from enum import IntEnum
from pathlib import Path
from typing import List

from pydantic import confloat, constr

from geolib.models.dseries_parser import DSeriesUnmappedNameProperties
from geolib.models.internal import Bool
from geolib.utils import csv_as_namedtuples

logger = logging.getLogger(__name__)


class SoilTypeNl(IntEnum):
    GRAVEL = 0
    SAND = 1
    LOAM = 2
    CLAY = 3
    PEAT = 4
    SANDYLOAM = 5


class SoilTypeBe(IntEnum):
    GRAVEL = 0
    SAND = 1
    LOAM = 2
    CLAY = 3
    PEAT = 4


class MaxConeResistType(IntEnum):
    STANDARD = 0
    MANUAL = 1


class Soil(DSeriesUnmappedNameProperties):
    name: constr(min_length=1, max_length=25)
    soilcolor: int = 10871211  # could be color
    soilsoiltype: SoilTypeNl = SoilTypeNl.SAND
    soilbelgiansoiltype: SoilTypeBe = SoilTypeBe.SAND
    soilgamdry: confloat(ge=0.0, le=100) = 20.00
    soilgamwet: confloat(ge=0.0, le=100) = 20.00
    soilinitialvoidratio: confloat(ge=0.0, le=20.0) = 0.001001
    soildiameterd50: confloat(ge=0.0, le=1000.0) = 0.20000
    soilminvoidratio: confloat(ge=0.0, le=1.0) = 0.400
    soilmaxvoidratio: confloat(ge=0.0, le=1.0) = 0.800
    soilcohesion: confloat(ge=0.0, le=1000.0) = 30.00
    soilphi: confloat(ge=0.0, le=89.0)
    soilcu: confloat(ge=0.0, le=1000.0) = 0.00
    soilmaxconeresisttype: MaxConeResistType = MaxConeResistType.STANDARD
    soilmaxconeresist: confloat(ge=0.0, le=1000000.0) = 0.00
    soilusetension: Bool = Bool.TRUE
    soilca: confloat(ge=0.0, le=10.0) = 0.0040000
    soilccindex: confloat(ge=0.0, le=20.0) = 0.1260000

    def __init__(self, *args, **kwargs):
        if "name" not in kwargs:
            name = None
            for key, value in kwargs.items():
                if key not in self.__fields__:
                    name = key + value
                    break
            if name:
                kwargs["name"] = name
                kwargs.pop(key)
        super().__init__(*args, **kwargs)

    @classmethod
    def default_soils(cls, model: str = "BEARING_PILES") -> List["Soil"]:
        currentfolder = Path(__file__).parent
        name = model.lower()
        filename = currentfolder / f"soil_csv/{name}_soils.csv"
        if not filename.exists():
            logger.warning(f"No default soils supported for {model}")
            return []

        soils = [
            Soil(
                name=name,
                soilsoiltype=soilsoiltype,
                soilgamdry=soilgamdry,
                soilgamwet=soilgamwet,
                soilphi=soilphi,
                soilcohesion=soilcohesion,
                soilcu=soilcu,
                soilccindex=soilccindex,
                soilca=soilca,
                soilinitialvoidratio=soilinitialvoidratio,
                soilcolor=soilcolor,
            )
            for name, soilsoiltype, soilgamdry, soilgamwet, soilphi, soilcohesion, soilcu, soilccindex, soilca, soilinitialvoidratio, soilcolor in csv_as_namedtuples(
                filename
            )
        ]
        return soils



from __future__ import annotations

import re
from typing import Any, Dict, List, Tuple

from geolib.models.dseries_parser import (
    DSerieListGroupNextStructure,
    DSerieParser,
    DSeriesInlineProperties,
    DSeriesStructure,
    get_line_property_key_value,
    get_line_property_value,
)


class DFoundationsEnumStructure(DSeriesStructure):
    @classmethod
    def parse_text(cls, text):
        parsed_structure = DSerieParser.parse_group_as_dict(text)
        kwargs = cls.remove_enum_explanation(parsed_structure)
        return cls(**kwargs)

    @staticmethod
    def remove_enum_explanation(structure: Dict[str, Any]) -> Dict[str, Any]:
        """Remove trailing explanations from values of `structure`
        0 : Mechanical qc required -> 0
        """
        updated_structure = {}
        for key, value in structure.items():
            if isinstance(value, str) and ":" in value:
                value = get_line_property_value(value, reversed_key=True)
            updated_structure[key] = value
        return updated_structure


class DFoundationsInlineProperties(DSeriesInlineProperties):
    @classmethod
    def get_property_key_value(cls, text: str, expected_property: str) -> Tuple[str, str]:
        """Returns the value content for a line of format:
        value : key || value = key

        Args:
            text (str): Text line with value-key

        Returns:
            str: Filtered value.
        """
        if text:
            # If the value is not empty, extract its value (assume it's on the first position)
            _, value = get_line_property_key_value(text, reversed_key=True)
            return expected_property, value
        return expected_property, text


class DFoundationsTableWrapper(DSeriesStructure):
    @classmethod
    def parse_text(cls, text):
        """Parses a Table wrapped in another group such as:
        [TABLE GROUP]
        [COLUMN INDICATION]
        A
        B
        [END OF COLUMN INDICATION]
        [GROUP DATA]
        1 1
        2 2
        [END OF GROUP DATA]
        [END OF TABLE GROUP]

        Args:
            text (str): Wrapped table group to parse.

        Returns:
            DSerieStructure: Parsed structure.
        """

        def split_line(text: str) -> List[str]:
            parts = re.split(" |\t", text.strip())
            return list(filter(lambda part: part != "", parts))

        table_text = list(DSerieParser.parse_list_group(text).values())[0]
        table_data = list(DSerieParser.parse_list_group(table_text).values())
        # Expected two groups (column_indication and data)
        keys = table_data[0].split("\n")
        values_dict_list = [
            dict(zip(keys, values))
            for values in map(split_line, table_data[1].split("\n"))
        ]
        collection_property_name = list(cls.__fields__.items())[0][0]
        return cls(**{collection_property_name: values_dict_list})


class DFoundationsCPTCollectionWrapper(DSerieListGroupNextStructure):
    @classmethod
    def group_delimiter(cls) -> str:
        return "[NEXT OF NUMBER OF CPTS]"



import abc
from typing import List

from pydantic import FilePath

from geolib.errors import NotConcreteError
from geolib.models.base_model_structure import BaseModelStructure


class BaseParser(abc.ABC):
    @property
    @abc.abstractmethod
    def suffix_list(self) -> List[str]:
        raise NotConcreteError

    @abc.abstractmethod
    def parse(self, filename: FilePath):
        raise NotConcreteError

    def can_parse(self, filename: FilePath) -> bool:
        """Verifies if a file or directory can be parsed with this instance.

        :param filename: File path to the file or directory to parse.
        :type filename: FilePath
        :return: True if it can be parsed, False otherwise.
        :rtype: bool
        """
        if not filename:
            return False

        return (filename.suffix == "" and filename.name in self.suffix_list) or (
            filename.suffix in self.suffix_list
        )


class BaseParserProvider(abc.ABC):
    """Basic class for parser providers."""

    @property
    @abc.abstractmethod
    def input_parsers(self) -> BaseParser:
        raise NotConcreteError

    @property
    @abc.abstractmethod
    def output_parsers(self) -> BaseParser:
        raise NotConcreteError

    @property
    @abc.abstractmethod
    def parser_name(self) -> str:
        raise NotConcreteError

    def parse(self, filename: FilePath) -> BaseModelStructure:
        for parser in self.input_parsers:
            if parser.can_parse(filename):
                return parser.parse(filename)

        for parser in self.output_parsers:
            if parser.can_parse(filename):
                return parser.parse(filename)
        raise ValueError(f"Unknown extension {filename.suffix} for {self.parser_name}.")



class GEOLibError(Exception):
    """Base GEOLib Exception class."""


class CalculationError(GEOLibError):
    """CalculationError with a status_code."""

    def __init__(self, status_code, message):
        self.status_code = status_code
        self.message = message


class ParserError(GEOLibError):
    """Base class for Parser errors."""


class GEOLibNotImplementedError(GEOLibError, NotImplementedError):
    """Base class for not implemented abstract classes."""


NotConcreteError = GEOLibNotImplementedError("Should be implemented in concrete class.")



import json
from collections import defaultdict
from datetime import date, datetime
from enum import Enum
from itertools import chain
from math import isfinite
from typing import Dict, Generator, List, Optional, Set, Tuple, Union

from pydantic import ValidationError, conlist, root_validator, validator

from geolib import __version__ as version
from geolib.geometry import Point
from geolib.models.base_model_structure import BaseModelStructure
from geolib.soils import Soil
from geolib.utils import snake_to_camel

from .dstability_validator import DStabilityValidator
from .utils import children


class DStabilityBaseModelStructure(BaseModelStructure):
    def dict(_, *args, **kwargs):
        data = super().dict(*args, **kwargs)
        return {
            k: "NaN" if isinstance(v, float) and not isfinite(v) else v
            for k, v in data.items()
        }


class AnalysisTypeEnum(Enum):
    BISHOP = "Bishop"
    BISHOP_BRUTE_FORCE = "BishopBruteForce"
    SPENCER = "Spencer"
    SPENCER_GENETIC = "SpencerGenetic"
    UPLIFT_VAN = "UpliftVan"
    UPLIFT_VAN_PARTICLE_SWARM = "UpliftVanParticleSwarm"


AnalysisType = AnalysisTypeEnum


class BishopSlipCircleResult(DStabilityBaseModelStructure):
    x: float
    z: float
    radius: float


class UpliftVanSlipCircleResult(DStabilityBaseModelStructure):
    x_left: float
    z_left: float
    x_right: float
    z_right: float
    z_tangent: float


class SpencerSlipPlaneResult(DStabilityBaseModelStructure):
    slipplane: List[Point]


class DStabilitySubStructure(DStabilityBaseModelStructure):
    @classmethod
    def structure_name(cls):
        class_name = cls.__name__
        return str.split(str.lower(class_name), ".")[-1]

    @classmethod
    def structure_group(cls):
        return cls.structure_name()


# waternet schema
class PersistablePoint(DStabilityBaseModelStructure):
    X: Optional[float] = "NaN"
    Z: Optional[float] = "NaN"


class PersistableHeadLine(DStabilityBaseModelStructure):
    Id: Optional[str]
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    Points: Optional[List[Optional[PersistablePoint]]]


class PersistableReferenceLine(DStabilityBaseModelStructure):
    BottomHeadLineId: Optional[str]
    Id: Optional[str]
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    Points: Optional[List[Optional[PersistablePoint]]]
    TopHeadLineId: Optional[str]


class Waternet(DStabilitySubStructure):
    """waternets/waternet_x.json."""

    @classmethod
    def structure_group(cls) -> str:
        return "waternets"

    @classmethod
    def structure_name(cls) -> str:
        return "waternets"

    Id: Optional[str]
    ContentVersion: Optional[str] = "1"
    PhreaticLineId: Optional[str]
    HeadLines: List[PersistableHeadLine] = []
    ReferenceLines: List[PersistableReferenceLine] = []
    UnitWeightWater: Optional[float] = 9.81

    def get_head_line(self, head_line_id: str) -> PersistableHeadLine:
        for head_line in self.HeadLines:
            if head_line.Id == head_line_id:
                return head_line

        raise ValueError(f"No headline with id {head_line_id} in model.")

    def get_reference_line(self, reference_line_id: str) -> PersistableReferenceLine:
        for reference_line in self.ReferenceLines:
            if reference_line.Id == reference_line_id:
                return reference_line

        raise ValueError(f"No referenceline with id {reference_line_id} in model.")

    def has_head_line_id(self, head_line_id: str) -> bool:
        return head_line_id in {head_line.Id for head_line in self.HeadLines}

    def add_head_line(
        self,
        head_line_id: str,
        label: str,
        notes: str,
        points: List[Point],
        is_phreatic_line: bool,
    ) -> PersistableHeadLine:
        head_line = PersistableHeadLine(Id=head_line_id, Label=label, Notes=notes)
        head_line.Points = [PersistablePoint(X=p.x, Z=p.z) for p in points]

        self.HeadLines.append(head_line)
        if is_phreatic_line:
            self.PhreaticLineId = head_line.Id

        return head_line

    def add_reference_line(
        self,
        reference_line_id: str,
        label: str,
        notes: str,
        points: List[Point],
        bottom_head_line_id: str,
        top_head_line_id: str,
    ) -> PersistableReferenceLine:
        reference_line = PersistableReferenceLine(
            Id=reference_line_id, Label=label, Notes=notes
        )
        reference_line.Points = [PersistablePoint(X=p.x, Z=p.z) for p in points]

        if not self.has_head_line_id(bottom_head_line_id):
            raise ValueError(
                f"Unknown headline id {bottom_head_line_id} for bottom_head_line_id"
            )

        if not self.has_head_line_id(top_head_line_id):
            raise ValueError(
                f"Unknown headline id {top_head_line_id} for top_head_line_id"
            )

        reference_line.BottomHeadLineId = bottom_head_line_id
        reference_line.TopHeadLineId = top_head_line_id

        self.ReferenceLines.append(reference_line)
        return reference_line


class PersistableDitchCharacteristics(DStabilityBaseModelStructure):
    DitchBottomEmbankmentSide: Optional[float] = "NaN"
    DitchBottomLandSide: Optional[float] = "NaN"
    DitchEmbankmentSide: Optional[float] = "NaN"
    DitchLandSide: Optional[float] = "NaN"


class PersistableEmbankmentCharacteristics(DStabilityBaseModelStructure):
    EmbankmentToeLandSide: Optional[float] = "NaN"
    EmbankmentToeWaterSide: Optional[float] = "NaN"
    EmbankmentTopLandSide: Optional[float] = "NaN"
    EmbankmentTopWaterSide: Optional[float] = "NaN"
    ShoulderBaseLandSide: Optional[float] = "NaN"


class EmbankmentSoilScenarioEnum(str, Enum):
    CLAY_EMBANKMENT_ON_CLAY = "ClayEmbankmentOnClay"
    CLAY_EMBANKMENT_ON_SAND = "ClayEmbankmentOnSand"
    SAND_EMBANKMENT_ON_CLAY = "SandEmbankmentOnClay"
    SAND_EMBANKMENT_ON_SAND = "SandEmbankmentOnSand"


class WaternetCreatorSettings(DStabilitySubStructure):
    """waternetcreatorsettings/waternetcreatorsettings_x.json"""

    AdjustForUplift: Optional[bool] = False
    AquiferInsideAquitardLayerId: Optional[str]
    AquiferLayerId: Optional[str]
    AquiferLayerInsideAquitardLeakageLengthInwards: Optional[float] = "NaN"
    AquiferLayerInsideAquitardLeakageLengthOutwards: Optional[float] = "NaN"
    AquitardHeadLandSide: Optional[float] = "NaN"
    AquitardHeadWaterSide: Optional[float] = "NaN"
    ContentVersion: Optional[str] = "1"
    DitchCharacteristics: Optional[
        PersistableDitchCharacteristics
    ] = PersistableDitchCharacteristics()
    DrainageConstruction: Optional[PersistablePoint] = PersistablePoint()
    EmbankmentCharacteristics: Optional[
        PersistableEmbankmentCharacteristics
    ] = PersistableEmbankmentCharacteristics()
    EmbankmentSoilScenario: EmbankmentSoilScenarioEnum = "ClayEmbankmentOnClay"
    Id: Optional[str]
    InitialLevelEmbankmentTopLandSide: Optional[float] = "NaN"
    InitialLevelEmbankmentTopWaterSide: Optional[float] = "NaN"
    IntrusionLength: Optional[float] = "NaN"
    IsAquiferLayerInsideAquitard: Optional[bool] = False
    IsDitchPresent: Optional[bool] = False
    IsDrainageConstructionPresent: Optional[bool] = False
    MeanWaterLevel: Optional[float] = "NaN"
    NormativeWaterLevel: Optional[float] = "NaN"
    OffsetEmbankmentToeLandSide: Optional[float] = "NaN"
    OffsetEmbankmentTopLandSide: Optional[float] = "NaN"
    OffsetEmbankmentTopWaterSide: Optional[float] = "NaN"
    OffsetShoulderBaseLandSide: Optional[float] = "NaN"
    PleistoceneLeakageLengthInwards: Optional[float] = "NaN"
    PleistoceneLeakageLengthOutwards: Optional[float] = "NaN"
    UseDefaultOffsets: Optional[bool] = True
    WaterLevelHinterland: Optional[float] = "NaN"

    @classmethod
    def structure_group(cls) -> str:
        return "waternetcreatorsettings"


class PersistableStochasticParameter(DStabilityBaseModelStructure):
    IsProbabilistic: bool = False
    Mean: float = 1.0
    StandardDeviation: float = 0.0


class StateType(Enum):
    OCR = "Ocr"
    POP = "Pop"
    YIELD_STRESS = "YieldStress"


class PersistableStress(DStabilityBaseModelStructure):
    Ocr: float = 1.0
    Pop: float = 0.0
    PopStochasticParameter: PersistableStochasticParameter = (
        PersistableStochasticParameter()
    )
    StateType: Optional[StateType]
    YieldStress: float = 0.0


class PersistableStateLinePoint(DStabilityBaseModelStructure):
    Above: Optional[PersistableStress]
    Below: Optional[PersistableStress]
    Id: Optional[str]
    IsAboveAndBelowCorrelated: Optional[bool]
    IsProbabilistic: Optional[bool]
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    X: Optional[float]


class PersistableStateLine(DStabilityBaseModelStructure):
    Points: Optional[List[Optional[PersistablePoint]]]
    Values: Optional[List[Optional[PersistableStateLinePoint]]]


class PersistableStatePoint(DStabilityBaseModelStructure):
    Id: Optional[str]
    IsProbabilistic: Optional[bool]
    Label: Optional[str] = ""
    LayerId: Optional[str]
    Notes: Optional[str] = ""
    Point: Optional[PersistablePoint]
    Stress: Optional[PersistableStress]


class State(DStabilitySubStructure):
    """states/states_x.json"""

    @classmethod
    def structure_name(cls) -> str:
        return "states"

    @classmethod
    def structure_group(cls) -> str:
        return "states"

    ContentVersion: Optional[str] = "1"
    Id: Optional[str]
    StateLines: List[PersistableStateLine] = []
    StatePoints: List[PersistableStatePoint] = []

    def add_state_point(self, state_point: PersistableStatePoint) -> None:
        self.StatePoints.append(state_point)

    def add_state_line(
        self,
        points: List[PersistablePoint],
        state_points: List[PersistableStateLinePoint],
    ):
        self.StateLines.append(PersistableStateLine(Points=points, Values=state_points))


# statecorrelation


class PersistableStateCorrelation(DStabilityBaseModelStructure):
    CorrelatedStateIds: Optional[List[Optional[str]]]
    IsFullyCorrelated: Optional[bool]


class StateCorrelation(DStabilitySubStructure):
    """statecorrelations/statecorrelations_1.json"""

    @classmethod
    def structure_name(cls) -> str:
        return "statecorrelations"

    @classmethod
    def structure_group(cls) -> str:
        return "statecorrelations"

    ContentVersion: Optional[str] = "1"
    Id: Optional[str]
    StateCorrelations: Optional[List[Optional[PersistableStateCorrelation]]] = []


class Stage(DStabilitySubStructure):
    """stages/stage_x.json"""

    @classmethod
    def structure_name(cls) -> str:
        return "stage"

    @classmethod
    def structure_group(cls) -> str:
        return "stages"

    AnalysisType: Optional[AnalysisTypeEnum] = AnalysisType.BISHOP_BRUTE_FORCE
    CalculationSettingsId: Optional[str]
    ContentVersion: Optional[str] = "1"
    DecorationsId: Optional[str]
    GeometryId: Optional[str]
    Id: Optional[str]
    Label: Optional[str] = ""
    LoadsId: Optional[str]
    Notes: Optional[str] = ""
    ReinforcementsId: Optional[str]
    ResultId: Optional[str] = None
    SoilLayersId: Optional[str]
    StateCorrelationsId: Optional[str]
    StateId: Optional[str]
    WaternetCreatorSettingsId: Optional[str]
    WaternetId: Optional[str]


class PersistableShadingType(Enum):
    DIAGONAL_A = "DiagonalA"
    DIAGONAL_B = "DiagonalB"
    DIAGONAL_C = "DiagonalC"
    DIAGONAL_D = "DiagonalD"
    DOT_A = "DotA"
    DOT_B = "DotB"
    DOT_C = "DotC"
    DOT_D = "DotD"
    HORIZONTAL_A = "HorizontalA"
    HORIZONTAL_B = "HorizontalB"
    NONE = "None"


class PersistableSoilVisualization(DStabilityBaseModelStructure):
    Color: Optional[str]
    PersistableShadingType: Optional[PersistableShadingType]
    SoilId: Optional[str]


class SoilVisualisation(DStabilityBaseModelStructure):
    ContentVersion: Optional[str] = "1"
    SoilVisualizations: Optional[List[Optional[PersistableSoilVisualization]]] = []

    @classmethod
    def structure_name(cls) -> str:
        return "soilvisualizations"


class PersistableSoilLayer(DStabilityBaseModelStructure):
    LayerId: Optional[str]
    SoilId: Optional[str]


class SoilLayerCollection(DStabilitySubStructure):
    """soillayers/soillayers_x.json"""

    @classmethod
    def structure_name(cls) -> str:
        return "soillayers"

    @classmethod
    def structure_group(cls) -> str:
        return "soillayers"

    ContentVersion: Optional[str] = "1"
    Id: Optional[str]
    SoilLayers: List[PersistableSoilLayer] = []

    def add_soillayer(self, layer_id: str, soil_id: str) -> PersistableSoilLayer:
        psl = PersistableSoilLayer(LayerId=layer_id, SoilId=soil_id)
        self.SoilLayers.append(psl)
        return psl

    def get_ids(self, exclude_soil_layer_id: Optional[int]) -> Set[str]:
        if exclude_soil_layer_id is not None:
            exclude_soil_layer_id = str(exclude_soil_layer_id)
        return {
            layer.LayerId
            for layer in self.SoilLayers
            if layer.LayerId != exclude_soil_layer_id
        }


class PersistableSoilCorrelation(DStabilityBaseModelStructure):
    CorrelatedSoilIds: Optional[List[str]]


class SoilCorrelation(DStabilitySubStructure):
    """soilcorrelations.json"""

    ContentVersion: Optional[str] = "1"
    SoilCorrelations: Optional[List[Optional[PersistableSoilCorrelation]]] = []

    @classmethod
    def structure_name(cls) -> str:
        return "soilcorrelations"


class ShearStrengthModelTypePhreaticLevelInternal(Enum):
    """
    Shear Strength Model Type. These types represent the types that
    are internally defined in the json files of D-Stability.
    """

    C_PHI = "CPhi"
    NONE = "None"
    SU = "Su"
    SUTABLE = "SuTable"

    def to_global_shear_strength_model(self):
        transform_dictionary = {
            "CPhi": "Mohr_Coulomb",
            "None": "None",
            "Su": "SHANSEP",
            "SuTable": "SuTable",
        }
        return transform_dictionary[self.value]


class PersistableSuTablePoint(DStabilitySubStructure):
    EffectiveStress: float = 0.0
    Su: float = 0.0


class PersistableSuTable(DStabilityBaseModelStructure):
    StrengthIncreaseExponent: float = 1.0
    StrengthIncreaseExponentStochasticParameter: PersistableStochasticParameter = (
        PersistableStochasticParameter()
    )
    SuTablePoints: List[PersistableSuTablePoint] = []
    IsSuTableProbabilistic: bool = False
    SuTableVariationCoefficient: float = 0.0

    def to_global_su_table(self):
        from geolib.soils import SuTablePoint

        su_table = []
        for su_table_point in self.SuTablePoints:
            su_table.append(
                SuTablePoint(su=su_table_point.Su, stress=su_table_point.EffectiveStress)
            )
        return su_table


class PersistableSoil(DStabilityBaseModelStructure):
    Code: str = ""
    Cohesion: float = 0.0
    CohesionAndFrictionAngleCorrelated: bool = False
    CohesionStochasticParameter: PersistableStochasticParameter = (
        PersistableStochasticParameter()
    )
    Dilatancy: float = 0.0
    DilatancyStochasticParameter: PersistableStochasticParameter = (
        PersistableStochasticParameter()
    )
    FrictionAngle: float = 0.0
    FrictionAngleStochasticParameter: PersistableStochasticParameter = (
        PersistableStochasticParameter()
    )
    Id: str = ""
    IsProbabilistic: bool = False
    Name: Optional[str] = ""
    Notes: Optional[str] = ""
    ShearStrengthModelTypeAbovePhreaticLevel: ShearStrengthModelTypePhreaticLevelInternal = (
        ShearStrengthModelTypePhreaticLevelInternal.C_PHI
    )
    ShearStrengthModelTypeBelowPhreaticLevel: ShearStrengthModelTypePhreaticLevelInternal = (
        ShearStrengthModelTypePhreaticLevelInternal.SU
    )
    ShearStrengthRatio: float = 0.0
    ShearStrengthRatioAndShearStrengthExponentCorrelated: bool = False
    ShearStrengthRatioStochasticParameter: PersistableStochasticParameter = (
        PersistableStochasticParameter()
    )
    StrengthIncreaseExponent: float = 1.0
    StrengthIncreaseExponentStochasticParameter: PersistableStochasticParameter = (
        PersistableStochasticParameter()
    )
    VolumetricWeightAbovePhreaticLevel: float = 0.0
    VolumetricWeightBelowPhreaticLevel: float = 0.0
    SuTable: PersistableSuTable = PersistableSuTable()


class SoilCollection(DStabilitySubStructure):
    """soils.json"""

    ContentVersion: Optional[str] = "1"
    Soils: List[PersistableSoil] = [
        PersistableSoil(
            Id="2",
            Name="Embankment new",
            Code="H_Aa_ht_new",
            Cohesion=7.0,
            FrictionAngle=30.0,
            Dilatancy=0.0,
            ShearStrengthRatio=0.26,
            StrengthIncreaseExponent=0.9,
            VolumetricWeightAbovePhreaticLevel=19.3,
            VolumetricWeightBelowPhreaticLevel=19.3,
        ),
        PersistableSoil(
            Id="3",
            Name="Embankment old",
            Code="H_Aa_ht_old",
            Cohesion=7.0,
            FrictionAngle=30.0,
            Dilatancy=0.0,
            ShearStrengthRatio=0.26,
            StrengthIncreaseExponent=0.9,
            VolumetricWeightAbovePhreaticLevel=18.0,
            VolumetricWeightBelowPhreaticLevel=18.0,
        ),
        PersistableSoil(
            Id="4",
            Name="Clay, shallow",
            Code="H_Rk_k_shallow",
            Cohesion=0.0,
            FrictionAngle=0.0,
            Dilatancy=0.0,
            ShearStrengthRatio=0.23,
            ShearStrengthModelTypeAbovePhreaticLevel=ShearStrengthModelTypePhreaticLevelInternal.SU,
            StrengthIncreaseExponent=0.9,
            VolumetricWeightAbovePhreaticLevel=14.8,
            VolumetricWeightBelowPhreaticLevel=14.8,
        ),
        PersistableSoil(
            Id="5",
            Name="Clay, deep",
            Code="H_Rk_k_deep",
            Cohesion=0.0,
            FrictionAngle=0.0,
            Dilatancy=0.0,
            ShearStrengthRatio=0.23,
            ShearStrengthModelTypeAbovePhreaticLevel=ShearStrengthModelTypePhreaticLevelInternal.SU,
            StrengthIncreaseExponent=0.9,
            VolumetricWeightAbovePhreaticLevel=15.6,
            VolumetricWeightBelowPhreaticLevel=15.6,
        ),
        PersistableSoil(
            Id="6",
            Name="Organic clay",
            Code="H_Rk_ko",
            Cohesion=0.0,
            FrictionAngle=0.0,
            Dilatancy=0.0,
            ShearStrengthRatio=0.24,
            ShearStrengthModelTypeAbovePhreaticLevel=ShearStrengthModelTypePhreaticLevelInternal.SU,
            StrengthIncreaseExponent=0.85,
            VolumetricWeightAbovePhreaticLevel=13.9,
            VolumetricWeightBelowPhreaticLevel=13.9,
        ),
        PersistableSoil(
            Id="7",
            Name="Peat, shallow",
            Code="H_vhv_v",
            Cohesion=0.0,
            FrictionAngle=0.0,
            Dilatancy=0.0,
            ShearStrengthRatio=0.3,
            ShearStrengthModelTypeAbovePhreaticLevel=ShearStrengthModelTypePhreaticLevelInternal.SU,
            StrengthIncreaseExponent=0.9,
            VolumetricWeightAbovePhreaticLevel=10.1,
            VolumetricWeightBelowPhreaticLevel=10.1,
        ),
        PersistableSoil(
            Id="8",
            Name="Peat, deep",
            Code="H_vbv_v",
            Cohesion=0.0,
            FrictionAngle=0.0,
            Dilatancy=0.0,
            ShearStrengthRatio=0.27,
            ShearStrengthModelTypeAbovePhreaticLevel=ShearStrengthModelTypePhreaticLevelInternal.SU,
            StrengthIncreaseExponent=0.9,
            VolumetricWeightAbovePhreaticLevel=11.0,
            VolumetricWeightBelowPhreaticLevel=11.0,
        ),
        PersistableSoil(
            Id="9",
            Name="Sand",
            Code="Sand",
            Cohesion=0.0,
            FrictionAngle=30.0,
            Dilatancy=0.0,
            ShearStrengthRatio=0.0,
            ShearStrengthModelTypeBelowPhreaticLevel=ShearStrengthModelTypePhreaticLevelInternal.C_PHI,
            StrengthIncreaseExponent=0.0,
            VolumetricWeightAbovePhreaticLevel=18.0,
            VolumetricWeightBelowPhreaticLevel=20.0,
        ),
        PersistableSoil(
            Id="10",
            Name="Clay with silt",
            Code="P_Rk_k&s",
            Cohesion=0.0,
            FrictionAngle=0.0,
            Dilatancy=0.0,
            ShearStrengthRatio=0.22,
            ShearStrengthModelTypeAbovePhreaticLevel=ShearStrengthModelTypePhreaticLevelInternal.SU,
            StrengthIncreaseExponent=0.9,
            VolumetricWeightAbovePhreaticLevel=18.0,
            VolumetricWeightBelowPhreaticLevel=18.0,
        ),
        PersistableSoil(
            Id="11",
            Name="Sand with clay",
            Code="H_Ro_z&k",
            Cohesion=0.0,
            FrictionAngle=0.0,
            Dilatancy=0.0,
            ShearStrengthRatio=0.22,
            ShearStrengthModelTypeAbovePhreaticLevel=ShearStrengthModelTypePhreaticLevelInternal.SU,
            StrengthIncreaseExponent=0.9,
            VolumetricWeightAbovePhreaticLevel=18.0,
            VolumetricWeightBelowPhreaticLevel=18.0,
        ),
    ]

    @classmethod
    def structure_name(cls) -> str:
        return "soils"

    def has_soilcode(self, code: str) -> bool:
        """
        Checks if the soilcode is available in the current soil list.

        Args:
            code (str): code of the soil

        Returns:
            bool: True if found, False if not
        """
        return code in {s.Code for s in self.Soils}

    def add_soil(self, soil: Soil) -> PersistableSoil:
        """
        Add a new soil to the model.

        Args:
            soil (Soil): a new soil

        Returns:
            None
        """
        ps = soil._to_dstability()

        self.Soils.append(ps)
        return ps

    @staticmethod
    def __to_global_stochastic_parameter(
        persistable_stochastic_parameter: PersistableStochasticParameter,
    ):
        from geolib.soils import StochasticParameter

        return StochasticParameter(
            is_probabilistic=persistable_stochastic_parameter.IsProbabilistic,
            mean=persistable_stochastic_parameter.Mean,
            standard_deviation=persistable_stochastic_parameter.StandardDeviation,
        )

    def __determine_strength_increase_exponent(self, persistable_soil: PersistableSoil):
        # shear increase exponent taken from persistable_soil.SuTable or just from persistable_soil
        if (
            persistable_soil.ShearStrengthModelTypeAbovePhreaticLevel.value == "Su"
            or persistable_soil.ShearStrengthModelTypeBelowPhreaticLevel.value == "Su"
        ):
            # SHANSEP model is selected so the StrengthIncreaseExponentStochasticParameter from persistable_soil should be used
            return self.__to_global_stochastic_parameter(
                persistable_soil.StrengthIncreaseExponentStochasticParameter
            )
        elif (
            persistable_soil.ShearStrengthModelTypeAbovePhreaticLevel.value == "SuTable"
            or persistable_soil.ShearStrengthModelTypeBelowPhreaticLevel.value
            == "SuTable"
        ):
            # SU table is selected so the StrengthIncreaseExponentStochasticParameter from SuTable should be used
            return self.__to_global_stochastic_parameter(
                persistable_soil.SuTable.StrengthIncreaseExponentStochasticParameter
            )
        else:
            return None

    def __internal_soil_to_global_soil(self, persistable_soil: PersistableSoil):
        from geolib.soils import (
            MohrCoulombParameters,
            SoilWeightParameters,
            UndrainedParameters,
        )

        mohr_coulomb_parameters = MohrCoulombParameters(
            cohesion=self.__to_global_stochastic_parameter(
                persistable_soil.CohesionStochasticParameter
            ),
            friction_angle=self.__to_global_stochastic_parameter(
                persistable_soil.FrictionAngleStochasticParameter
            ),
            dilatancy_angle=self.__to_global_stochastic_parameter(
                persistable_soil.DilatancyStochasticParameter
            ),
            cohesion_and_friction_angle_correlated=persistable_soil.CohesionAndFrictionAngleCorrelated,
        )

        strength_increase_exponent = self.__determine_strength_increase_exponent(
            persistable_soil
        )
        undrained_parameters = UndrainedParameters(
            shear_strength_ratio=self.__to_global_stochastic_parameter(
                persistable_soil.ShearStrengthRatioStochasticParameter
            ),
            strength_increase_exponent=strength_increase_exponent,
            shear_strength_ratio_and_shear_strength_exponent_correlated=persistable_soil.ShearStrengthRatioAndShearStrengthExponentCorrelated,
            su_table=persistable_soil.SuTable.to_global_su_table(),
            probabilistic_su_table=persistable_soil.SuTable.IsSuTableProbabilistic,
            su_table_variation_coefficient=persistable_soil.SuTable.SuTableVariationCoefficient,
        )

        soil_weight_parameters = SoilWeightParameters()
        soil_weight_parameters.saturated_weight.mean = (
            persistable_soil.VolumetricWeightAbovePhreaticLevel
        )
        soil_weight_parameters.unsaturated_weight.mean = (
            persistable_soil.VolumetricWeightAbovePhreaticLevel
        )

        return Soil(
            id=persistable_soil.Id,
            name=persistable_soil.Name,
            code=persistable_soil.Code,
            is_probabilistic=persistable_soil.IsProbabilistic,
            shear_strength_model_above_phreatic_level=persistable_soil.ShearStrengthModelTypeAbovePhreaticLevel.to_global_shear_strength_model(),
            shear_strength_model_below_phreatic_level=persistable_soil.ShearStrengthModelTypeBelowPhreaticLevel.to_global_shear_strength_model(),
            mohr_coulomb_parameters=mohr_coulomb_parameters,
            soil_weight_parameters=soil_weight_parameters,
            undrained_parameters=undrained_parameters,
        )

    def get_soil(self, code: str) -> Soil:
        """
        Get soil by the given code.

        Args:
            code (str): code of the soil

        Returns:
            Soil: the soil object
        """
        for persistable_soil in self.Soils:
            if persistable_soil.Code == code:
                return self.__internal_soil_to_global_soil(persistable_soil)

        raise ValueError(f"Soil code '{code}' not found in the SoilCollection")

    def edit_soil(self, code: str, **kwargs: dict) -> PersistableSoil:
        """
        Update a soil.

        Args:
            code (str): code of the soil
            kwargs (dict): dictionary with agument names and values

        Returns:
            PersistableSoil: the edited soil
        """
        for persistable_soil in self.Soils:
            if persistable_soil.Code == code:
                for k, v in kwargs.items():
                    try:
                        setattr(persistable_soil, snake_to_camel(k), v)

                        k_stochastic = f"{snake_to_camel(k)}StochasticParameter"
                        if hasattr(persistable_soil, k_stochastic):
                            getattr(persistable_soil, k_stochastic).Mean = v
                    except AttributeError:
                        raise ValueError(f"Unknown soil parameter {k}.")

                return persistable_soil

        raise ValueError(f"Soil code '{code}' not found in the SoilCollection")


# Reinforcements


class PersistableForbiddenLine(DStabilityBaseModelStructure):
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    End: Optional[PersistablePoint]
    Start: Optional[PersistablePoint]


class PersistableGeotextile(DStabilityBaseModelStructure):
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    End: Optional[PersistablePoint]
    ReductionArea: Optional[float]
    Start: Optional[PersistablePoint]
    TensileStrength: Optional[float]


class PersistableStressAtDistance(DStabilityBaseModelStructure):
    Distance: Optional[float]
    Stress: Optional[float]


class PersistableNail(DStabilityBaseModelStructure):
    BendingStiffness: Optional[float] = 0.0
    CriticalAngle: Optional[float] = 0.0
    Diameter: Optional[float]
    Direction: Optional[float] = 0.0
    GroutDiameter: Optional[float] = 0.0
    HorizontalSpacing: Optional[float] = 0.0
    Label: Optional[str] = ""
    LateralStresses: Optional[List[Optional[PersistableStressAtDistance]]] = []
    Length: Optional[float]
    Location: Optional[PersistablePoint]
    MaxPullForce: Optional[float] = 0.0
    Notes: Optional[str] = ""
    PlasticMoment: Optional[float] = 0.0
    ShearStresses: Optional[List[Optional[PersistableStressAtDistance]]] = []
    UseFacing: Optional[bool] = False
    UseLateralStress: Optional[bool] = False
    UseShearStress: Optional[bool] = False


class Reinforcements(DStabilitySubStructure):
    """reinforcements/reinforcements_x.json"""

    Id: Optional[str]
    ContentVersion: Optional[str] = "1"
    ForbiddenLines: List[PersistableForbiddenLine] = []
    Geotextiles: List[PersistableGeotextile] = []
    Nails: List[PersistableNail] = []

    def add_reinforcement(
        self, reinforcement: "DStabilityReinforcement"
    ) -> Union[PersistableForbiddenLine, PersistableGeotextile, PersistableNail]:
        internal_datastructure = reinforcement._to_internal_datastructure()
        plural_class_name = f"{reinforcement.__class__.__name__}s"
        getattr(self, plural_class_name).append(internal_datastructure)
        return internal_datastructure


class ProjectInfo(DStabilitySubStructure):
    """projectinfo.json."""

    Analyst: Optional[str] = ""
    ApplicationCreated: Optional[str] = ""
    ApplicationModified: Optional[str] = ""
    ContentVersion: Optional[str] = "1"
    Created: Optional[date] = datetime.now().date()
    CrossSection: Optional[str] = ""
    Date: Optional[date] = datetime.now().date()
    IsDataValidated: Optional[bool] = False
    LastModified: Optional[date] = datetime.now().date()
    LastModifier: Optional[str] = "GEOLib"
    Path: Optional[str] = ""
    Project: Optional[str] = ""
    Remarks: Optional[str] = f"Created with GEOLib {version}"

    @validator("Created", "Date", "LastModified", pre=True, allow_reuse=True)
    def nltime(cls, date: Union[date, str]) -> date:
        if isinstance(date, str):
            position = date.index(max(date.split("-"), key=len))
            if position > 0:
                date = datetime.strptime(date, "%d-%m-%Y").date()
            else:
                date = datetime.strptime(date, "%Y-%m-%d").date()
        return date


class PersistableBondStress(DStabilityBaseModelStructure):
    Sigma: Optional[float]
    Tau: Optional[float]


class PersistableNailPropertiesForSoil(DStabilityBaseModelStructure):
    AreBondStressesActive: Optional[bool] = False
    BondStresses: Optional[List[Optional[PersistableBondStress]]] = []
    CompressionRatio: Optional[float]
    RheologicalCoefficient: Optional[float]
    SoilId: Optional[str]


class NailProperties(DStabilitySubStructure):
    """nailpropertiesforsoils.json"""

    ContentVersion: Optional[str] = "1"
    NailPropertiesForSoils: Optional[
        List[Optional[PersistableNailPropertiesForSoil]]
    ] = []

    @classmethod
    def structure_name(cls) -> str:
        return "nailpropertiesforsoils"


class PersistableConsolidation(DStabilityBaseModelStructure):
    Degree: Optional[float]
    LayerId: Optional[str]


class PersistableEarthquake(DStabilityBaseModelStructure):
    Consolidations: Optional[List[Optional[PersistableConsolidation]]] = []
    FreeWaterFactor: Optional[float] = 0.0
    HorizontalFactor: Optional[float] = 0.0
    IsEnabled: Optional[bool] = False
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    VerticalFactor: Optional[float] = 0.0


class PersistableLayerLoad(DStabilityBaseModelStructure):
    Consolidations: Optional[List[Optional[PersistableConsolidation]]] = []
    LayerId: Optional[str]


class PersistableLineLoad(DStabilityBaseModelStructure):
    Angle: Optional[float]
    Consolidations: Optional[List[Optional[PersistableConsolidation]]] = []
    Label: Optional[str] = ""
    Location: Optional[PersistablePoint]
    Magnitude: Optional[float]
    Notes: Optional[str] = ""
    Spread: Optional[float]


class PersistableTree(DStabilityBaseModelStructure):
    Force: Optional[float]
    Label: Optional[str] = ""
    Location: Optional[PersistablePoint]
    Notes: Optional[str] = ""
    RootZoneWidth: Optional[float]
    Spread: Optional[float]


class PersistableUniformLoad(DStabilityBaseModelStructure):
    Consolidations: Optional[List[Optional[PersistableConsolidation]]] = []
    End: Optional[float]
    Label: Optional[str] = ""
    Magnitude: Optional[float]
    Notes: Optional[str] = ""
    Spread: Optional[float]
    Start: Optional[float]


Load = Union[PersistableUniformLoad, PersistableLineLoad, PersistableLayerLoad]


class Loads(DStabilitySubStructure):
    """loads/loads_x.json"""

    Id: Optional[str]
    ContentVersion: Optional[str] = "1"
    Earthquake: Optional[PersistableEarthquake] = PersistableEarthquake()
    LayerLoads: Optional[List[Optional[PersistableLayerLoad]]] = []
    LineLoads: Optional[List[Optional[PersistableLineLoad]]] = []
    Trees: Optional[List[Optional[PersistableTree]]] = []
    UniformLoads: Optional[List[Optional[PersistableUniformLoad]]] = []

    def add_load(
        self, load: "DStabilityLoad", consolidations: List["Consolidation"]
    ) -> Union[PersistableUniformLoad, PersistableLineLoad, PersistableLayerLoad]:
        internal_datastructure = load.to_internal_datastructure()

        # Add consolidations if the load supports it
        if hasattr(internal_datastructure, "Consolidations"):
            internal_datastructure.Consolidations = [
                c.to_internal_datastructure() for c in consolidations
            ]

        target = load.__class__.__name__
        if target == "Earthquake":
            setattr(self, target, internal_datastructure)
        else:
            target += "s"
            getattr(self, target).append(internal_datastructure)

        return internal_datastructure

    def add_layer_load(
        self, soil_layer_id: int, consolidations: List["Consolidation"]
    ) -> PersistableLayerLoad:
        layer_load = PersistableLayerLoad(
            LayerId=str(soil_layer_id),
            Consolidations=[c.to_internal_datastructure() for c in consolidations],
        )
        self.LayerLoads.append(layer_load)
        return layer_load


class PersistableLayer(DStabilityBaseModelStructure):
    Id: Optional[str]
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    Points: conlist(PersistablePoint, min_items=3)

    @validator("Points", pre=True, allow_reuse=True)
    def polygon_checks(cls, points):
        """
        Todo:
            Find a way to check the validity of the given points
        """
        # implement some checks
        # 1. is this a simple polygon
        # 2. is it clockwise
        # 3. is it a non closed polygon
        # 4. does it intersect other polygons
        return points


class Geometry(DStabilitySubStructure):
    """geometries/geometry_x.json"""

    @classmethod
    def structure_group(cls) -> str:
        return "geometries"

    ContentVersion: Optional[str] = "1"
    Id: Optional[str]
    Layers: List[PersistableLayer] = []

    def contains_point(self, point: Point) -> bool:
        """
        Check if the given point is on one of the points of the layers

        Args:
            point (Point): A point type

        Returns:
            bool: True if this point is found on a layer, False otherwise

        Todo:
            Take x, z accuracy into account
        """
        for layer in self.Layers:
            for p in layer.Points:
                if point.x == p.X and point.z == p.Z:  # not nice
                    return True

        return False

    def get_layer(self, id: int) -> PersistableLayer:
        for layer in self.Layers:
            if layer.Id == str(id):
                return layer

        raise ValueError(f"Layer id {id} not found in this geometry")

    def add_layer(
        self, id: str, label: str, notes: str, points: List[Point]
    ) -> PersistableLayer:
        """
        Add a new layer to the model. Layers are expected;
        1. to contain at least 3 point (non closed polygons)
        2. the points need to be in clockwise order
        3. the polygon needs to be convex (no intersections with itsself)

        Args:
            id (str): id of the layer
            label (str): label of the layer
            notes (str): notes for the layers
            points (List[Points]): list of Point classes

        Returns:
            PersistableLayer: the layer as a persistable object
        """
        layer = PersistableLayer(
            Id=id,
            Label=label,
            Notes=notes,
            Points=[PersistablePoint(X=p.x, Z=p.z) for p in points],
        )

        self.Layers.append(layer)
        return layer


class PersistableBerm(DStabilityBaseModelStructure):
    AddedLayerId: Optional[str]
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    Points: Optional[List[Optional[PersistablePoint]]]


class PersistableExcavation(DStabilityBaseModelStructure):
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    Points: Optional[List[Optional[PersistablePoint]]]


class Decorations(DStabilitySubStructure):
    """decorations/decorations_x.json."""

    Berms: Optional[List[Optional[PersistableBerm]]] = []
    ContentVersion: Optional[str] = "1"
    Excavations: Optional[List[Optional[PersistableExcavation]]] = []
    Id: Optional[str]


# Calculation Settings


class PersistableCircle(DStabilityBaseModelStructure):
    Center: Optional[PersistablePoint] = PersistablePoint()
    Radius: Optional[float] = "NaN"


class PersistableBishopSettings(DStabilityBaseModelStructure):
    Circle: Optional[PersistableCircle] = PersistableCircle()
    Label: Optional[str] = ""
    Notes: Optional[str] = ""


class PersistableGridEnhancements(DStabilityBaseModelStructure):
    ExtrapolateSearchSpace: Optional[bool] = True


class NullablePersistablePoint(DStabilityBaseModelStructure):
    X: Optional[float] = "NaN"
    Z: Optional[float] = "NaN"


class PersistableSearchGrid(DStabilityBaseModelStructure):
    BottomLeft: Optional[NullablePersistablePoint] = None
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    NumberOfPointsInX: Optional[int] = 1
    NumberOfPointsInZ: Optional[int] = 1
    Space: Optional[float] = 1.0


class PersistableSlipPlaneConstraints(DStabilityBaseModelStructure):
    IsSizeConstraintsEnabled: Optional[bool] = False
    IsZoneAConstraintsEnabled: Optional[bool] = False
    IsZoneBConstraintsEnabled: Optional[bool] = False
    MinimumSlipPlaneDepth: Optional[float] = 0.0
    MinimumSlipPlaneLength: Optional[float] = 0.0
    WidthZoneA: Optional[float] = 0.0
    WidthZoneB: Optional[float] = 0.0
    XLeftZoneA: Optional[float] = 0.0
    XLeftZoneB: Optional[float] = 0.0


class PersistableTangentLines(DStabilityBaseModelStructure):
    BottomTangentLineZ: Optional[float] = "NaN"
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    NumberOfTangentLines: Optional[int] = 1
    Space: Optional[float] = 0.5


class PersistableBishopBruteForceSettings(DStabilityBaseModelStructure):
    GridEnhancements: Optional[
        PersistableGridEnhancements
    ] = PersistableGridEnhancements()
    SearchGrid: Optional[PersistableSearchGrid] = PersistableSearchGrid()
    SlipPlaneConstraints: Optional[
        PersistableSlipPlaneConstraints
    ] = PersistableSlipPlaneConstraints()
    TangentLines: Optional[PersistableTangentLines] = PersistableTangentLines()


class CalculationTypeEnum(Enum):
    DESIGN = "Design"
    DETERMINISTIC = "Deterministic"
    MEAN = "Mean"
    PROBABILISTIC = "Probabilistic"


CalculationType = CalculationTypeEnum


class PersistableSpencerSettings(DStabilityBaseModelStructure):
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    SlipPlane: Optional[List[Optional[PersistablePoint]]] = None


class OptionsTypeEnum(Enum):
    DEFAULT = "Default"
    THOROUGH = "Thorough"


OptionsType = OptionsTypeEnum


class PersistableGeneticSlipPlaneConstraints(DStabilityBaseModelStructure):
    IsEnabled: Optional[bool] = False
    MinimumAngleBetweenSlices: Optional[float] = 0.0
    MinimumThrustLinePercentageInsideSlices: Optional[float] = 0.0


class PersistableSpencerGeneticSettings(DStabilityBaseModelStructure):
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    OptionsType: Optional[OptionsTypeEnum] = OptionsType.DEFAULT
    SlipPlaneA: Optional[List[Optional[PersistablePoint]]] = None
    SlipPlaneB: Optional[List[Optional[PersistablePoint]]] = None
    SlipPlaneConstraints: Optional[
        PersistableGeneticSlipPlaneConstraints
    ] = PersistableGeneticSlipPlaneConstraints()


class PersistableTwoCirclesOnTangentLine(DStabilityBaseModelStructure):
    FirstCircleCenter: Optional[NullablePersistablePoint] = NullablePersistablePoint()
    FirstCircleRadius: Optional[float] = "NaN"
    SecondCircleCenter: Optional[NullablePersistablePoint] = NullablePersistablePoint()


class PersistableUpliftVanSettings(DStabilityBaseModelStructure):
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    SlipPlane: Optional[
        PersistableTwoCirclesOnTangentLine
    ] = PersistableTwoCirclesOnTangentLine()


class PersistableSearchArea(DStabilityBaseModelStructure):
    Height: Optional[float] = 0.0
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    TopLeft: Optional[NullablePersistablePoint] = None
    Width: Optional[float] = 0.0


class PersistableTangentArea(DStabilityBaseModelStructure):
    Height: Optional[float] = 0.0
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    TopZ: Optional[float] = None


class PersistableUpliftVanParticleSwarmSettings(DStabilityBaseModelStructure):
    Label: Optional[str] = ""
    Notes: Optional[str] = ""
    OptionsType: Optional[OptionsTypeEnum] = OptionsType.DEFAULT
    SearchAreaA: Optional[PersistableSearchArea] = PersistableSearchArea()
    SearchAreaB: Optional[PersistableSearchArea] = PersistableSearchArea()
    SlipPlaneConstraints: Optional[
        PersistableSlipPlaneConstraints
    ] = PersistableSlipPlaneConstraints()
    TangentArea: Optional[PersistableTangentArea] = PersistableTangentArea()


class CalculationSettings(DStabilitySubStructure):
    """calculationsettings/calculationsettings_x.json"""

    AnalysisType: Optional[AnalysisTypeEnum]
    Bishop: Optional[PersistableBishopSettings] = PersistableBishopSettings()
    BishopBruteForce: Optional[
        PersistableBishopBruteForceSettings
    ] = PersistableBishopBruteForceSettings()
    CalculationType: Optional[CalculationTypeEnum] = CalculationTypeEnum.DETERMINISTIC
    ContentVersion: Optional[str] = "1"
    Id: Optional[str] = "19"
    ModelFactorMean: Optional[float] = 1.05
    ModelFactorStandardDeviation: Optional[float] = 0.033
    Spencer: Optional[PersistableSpencerSettings] = PersistableSpencerSettings()
    SpencerGenetic: Optional[
        PersistableSpencerGeneticSettings
    ] = PersistableSpencerGeneticSettings()
    UpliftVan: Optional[PersistableUpliftVanSettings] = PersistableUpliftVanSettings()
    UpliftVanParticleSwarm: Optional[
        PersistableUpliftVanParticleSwarmSettings
    ] = PersistableUpliftVanParticleSwarmSettings()

    def set_bishop(self, bishop_settings: PersistableBishopSettings) -> None:
        self.Bishop = bishop_settings
        self.AnalysisType = AnalysisType.BISHOP

    def set_bishop_brute_force(
        self, bishop_brute_force_settings: PersistableBishopBruteForceSettings
    ) -> None:
        self.BishopBruteForce = bishop_brute_force_settings
        self.AnalysisType = AnalysisType.BISHOP_BRUTE_FORCE

    def set_spencer(self, spencer_settings: PersistableSpencerSettings) -> None:
        self.Spencer = spencer_settings
        self.AnalysisType = AnalysisType.SPENCER

    def set_spencer_genetic(
        self, spencer_genetic_settings: PersistableSpencerGeneticSettings
    ) -> None:
        self.SpencerGenetic = spencer_genetic_settings
        self.AnalysisType = AnalysisType.SPENCER_GENETIC

    def set_uplift_van(self, uplift_van_settings: PersistableUpliftVanSettings) -> None:
        self.UpliftVan = uplift_van_settings
        self.AnalysisType = AnalysisType.UPLIFT_VAN

    def set_uplift_van_particle_swarm(
        self,
        uplift_van_particle_swarm_settings: PersistableUpliftVanParticleSwarmSettings,
    ) -> None:
        self.UpliftVanParticleSwarm = uplift_van_particle_swarm_settings
        self.AnalysisType = AnalysisType.UPLIFT_VAN_PARTICLE_SWARM


########
# OUTPUT
########


class PersistableSlice(DStabilityBaseModelStructure):
    ArcLength: Optional[float] = None
    BottomAngle: Optional[float] = None
    BottomLeft: Optional[PersistablePoint] = None
    BottomRight: Optional[PersistablePoint] = None
    CohesionInput: Optional[float] = None
    CohesionOutput: Optional[float] = None
    DegreeOfConsolidationLoadPorePressure: Optional[float] = None
    DegreeOfConsolidationPorePressure: Optional[float] = None
    DilatancyInput: Optional[float] = None
    DilatancyOutput: Optional[float] = None
    EffectiveStress: Optional[float] = None
    HorizontalPorePressure: Optional[float] = None
    HorizontalSoilQuakeStress: Optional[float] = None
    HydrostaticPorePressure: Optional[float] = None
    Label: Optional[str] = None
    LoadStress: Optional[float] = None
    MInput: Optional[float] = None
    NormalStress: Optional[float] = None
    Ocr: Optional[float] = None
    PhiInput: Optional[float] = None
    PhiOutput: Optional[float] = None
    PiezometricPorePressure: Optional[float] = None
    Pop: Optional[float] = None
    ShearStress: Optional[float] = None
    SInput: Optional[float] = None
    SuOutput: Optional[float] = None
    SurfacePorePressure: Optional[float] = None
    TopAngle: Optional[float] = None
    TopLeft: Optional[PersistablePoint] = None
    TopRight: Optional[PersistablePoint] = None
    TotalPorePressure: Optional[float] = None
    TotalStress: Optional[float] = None
    UpliftFactor: Optional[float] = None
    VerticalPorePressure: Optional[float] = None
    VerticalSoilQuakeStress: Optional[float] = None
    WaterQuakeStress: Optional[float] = None
    Weight: Optional[float] = None
    Width: Optional[float] = None
    YieldStress: Optional[float] = None
    ShearStrengthModelType: Optional[ShearStrengthModelTypePhreaticLevelInternal] = None


class BishopBruteForceResult(DStabilitySubStructure):
    Circle: Optional[PersistableCircle] = None
    FactorOfSafety: Optional[float] = None
    Id: Optional[str] = None
    Points: Optional[List[Optional[PersistablePoint]]] = None
    Slices: Optional[List[Optional[PersistableSlice]]] = None

    @classmethod
    def structure_group(cls) -> str:
        return "results/bishopbruteforce/"

    def get_slipcircle_output(self) -> BishopSlipCircleResult:
        """Get condensed slipcircle data"""
        try:
            return BishopSlipCircleResult(
                x=self.Circle.Center.X, z=self.Circle.Center.Z, radius=self.Circle.Radius
            )
        except (ValidationError, AttributeError):
            raise ValueError(
                f"Slipcircle not available for {self.__class__.__name__} with id {self.Id}"
            )


class PersistableSoilContribution(DStabilityBaseModelStructure):
    Alpha: Optional[float] = None
    Property: Optional[str] = None
    SoilId: Optional[str] = None
    UncorrelatedAlpha: Optional[float] = None
    Value: Optional[float] = None


class PersistableStageContribution(DStabilityBaseModelStructure):
    Alpha: Optional[float] = None
    Property: Optional[str] = None
    StageId: Optional[str] = None
    UncorrelatedAlpha: Optional[float] = None
    Value: Optional[float] = None


class PersistableStateLinePointContribution(DStabilityBaseModelStructure):
    Alpha: Optional[float] = None
    Property: Optional[str] = None
    StateLinePointId: Optional[str] = None
    UncorrelatedAlpha: Optional[float] = None
    Value: Optional[float] = None


class PersistableStatePointContribution(DStabilityBaseModelStructure):
    Alpha: Optional[float] = None
    Property: Optional[str] = None
    StatePointId: Optional[str] = None
    UncorrelatedAlpha: Optional[float] = None
    Value: Optional[float] = None


class BishopReliabilityResult(DStabilitySubStructure):
    Circle: Optional[PersistableCircle] = None
    Converged: Optional[bool] = None
    FailureProbability: Optional[float] = None
    Id: Optional[str] = None
    ReliabilityIndex: Optional[float] = None
    SoilContributions: Optional[List[Optional[PersistableSoilContribution]]] = None
    StageContributions: Optional[List[Optional[PersistableStageContribution]]] = None
    StateLinePointContributions: Optional[
        List[Optional[PersistableStateLinePointContribution]]
    ] = None
    StatePointContributions: Optional[
        List[Optional[PersistableStatePointContribution]]
    ] = None

    @classmethod
    def structure_group(cls) -> str:
        return "results/bishopreliability/"

    def get_slipcircle_output(self) -> BishopSlipCircleResult:
        """Get condensed slipcircle data"""
        try:
            return BishopSlipCircleResult(
                x=self.Circle.Center.X, z=self.Circle.Center.Z, radius=self.Circle.Radius
            )
        except (ValidationError, AttributeError):
            raise ValueError(
                f"Slipcircle not available for {self.__class__.__name__} with id {self.Id}"
            )


class BishopResult(DStabilitySubStructure):
    Circle: Optional[PersistableCircle] = None
    FactorOfSafety: Optional[float] = None
    Id: Optional[str] = None
    Points: Optional[List[Optional[PersistablePoint]]] = None
    Slices: Optional[List[Optional[PersistableSlice]]] = None

    @classmethod
    def structure_group(cls) -> str:
        return "results/bishop/"

    def get_slipcircle_output(self) -> BishopSlipCircleResult:
        """Get condensed slipcircle data"""
        try:
            return BishopSlipCircleResult(
                x=self.Circle.Center.X, z=self.Circle.Center.Z, radius=self.Circle.Radius
            )
        except (ValidationError, AttributeError):
            raise ValueError(
                f"Slipcircle not available for {self.__class__.__name__} with id {self.Id}"
            )


class PersistableSpencerSlice(DStabilityBaseModelStructure):
    ArcLength: Optional[float] = None
    BottomAngle: Optional[float] = None
    BottomLeft: Optional[PersistablePoint] = None
    BottomRight: Optional[PersistablePoint] = None
    CohesionInput: Optional[float] = None
    CohesionOutput: Optional[float] = None
    DegreeOfConsolidationLoadPorePressure: Optional[float] = None
    DegreeOfConsolidationPorePressure: Optional[float] = None
    DilatancyInput: Optional[float] = None
    DilatancyOutput: Optional[float] = None
    EffectiveStress: Optional[float] = None
    HorizontalPorePressure: Optional[float] = None
    HorizontalSoilQuakeStress: Optional[float] = None
    HydrostaticPorePressure: Optional[float] = None
    Label: Optional[str] = None
    LeftForce: Optional[float] = None
    LeftForceAngle: Optional[float] = None
    LeftForceY: Optional[float] = None
    LoadStress: Optional[float] = None
    MInput: Optional[float] = None
    NormalStress: Optional[float] = None
    Ocr: Optional[float] = None
    PhiInput: Optional[float] = None
    PhiOutput: Optional[float] = None
    PiezometricPorePressure: Optional[float] = None
    Pop: Optional[float] = None
    RightForce: Optional[float] = None
    RightForceAngle: Optional[float] = None
    RightForceY: Optional[float] = None
    ShearStress: Optional[float] = None
    SInput: Optional[float] = None
    SuOutput: Optional[float] = None
    SurfacePorePressure: Optional[float] = None
    TopAngle: Optional[float] = None
    TopLeft: Optional[PersistablePoint] = None
    TopRight: Optional[PersistablePoint] = None
    TotalPorePressure: Optional[float] = None
    TotalStress: Optional[float] = None
    UpliftFactor: Optional[float] = None
    VerticalPorePressure: Optional[float] = None
    VerticalSoilQuakeStress: Optional[float] = None
    WaterQuakeStress: Optional[float] = None
    Weight: Optional[float] = None
    Width: Optional[float] = None
    YieldStress: Optional[float] = None
    ShearStrengthModelType: Optional[ShearStrengthModelTypePhreaticLevelInternal] = None


class SpencerGeneticAlgorithmResult(DStabilitySubStructure):
    FactorOfSafety: Optional[float] = None
    Id: Optional[str] = None
    Points: Optional[List[Optional[PersistablePoint]]] = None
    Slices: Optional[List[Optional[PersistableSpencerSlice]]] = None
    SlipPlane: Optional[List[Optional[PersistablePoint]]] = None

    @classmethod
    def structure_group(cls) -> str:
        return "results/spencergeneticalgorithm/"

    def get_slipplane_output(self) -> SpencerSlipPlaneResult:
        """Get condensed slipplane data"""
        try:
            return SpencerSlipPlaneResult(
                slipplane=[Point(x=p.X, z=p.Z) for p in self.SlipPlane]
            )
        except (ValidationError, TypeError):
            raise ValueError(
                f"Slip plane not available for {self.__class__.__name__} with id {self.Id}"
            )


class SpencerReliabilityResult(DStabilitySubStructure):
    Converged: Optional[bool] = None
    FailureProbability: Optional[float] = None
    Id: Optional[str] = None
    ReliabilityIndex: Optional[float] = None
    SlipPlane: Optional[List[Optional[PersistablePoint]]] = None
    SoilContributions: Optional[List[Optional[PersistableSoilContribution]]] = None
    StageContributions: Optional[List[Optional[PersistableStageContribution]]] = None
    StateLinePointContributions: Optional[
        List[Optional[PersistableStateLinePointContribution]]
    ] = None
    StatePointContributions: Optional[
        List[Optional[PersistableStatePointContribution]]
    ] = None

    @classmethod
    def structure_group(cls) -> str:
        return "results/spencerreliability/"

    def get_slipplane_output(self) -> SpencerSlipPlaneResult:
        """Get condensed slipplane data"""
        try:
            return SpencerSlipPlaneResult(
                slipplane=[Point(x=p.X, z=p.Z) for p in self.SlipPlane]
            )
        except (ValidationError, TypeError):
            raise ValueError(
                f"Slip plane not available for {self.__class__.__name__} with id {self.Id}"
            )


class SpencerResult(DStabilitySubStructure):
    FactorOfSafety: Optional[float] = None
    Id: Optional[str] = None
    Points: Optional[List[Optional[PersistablePoint]]] = None
    Slices: Optional[List[Optional[PersistableSpencerSlice]]] = None
    SlipPlane: Optional[List[Optional[PersistablePoint]]] = None

    @classmethod
    def structure_group(cls) -> str:
        return "results/spencer/"

    def get_slipplane_output(self) -> SpencerSlipPlaneResult:
        """Get condensed slipplane data"""
        try:
            return SpencerSlipPlaneResult(
                slipplane=[Point(x=p.X, z=p.Z) for p in self.SlipPlane]
            )
        except (ValidationError, TypeError):
            raise ValueError(
                f"Slip plane not available for {self.__class__.__name__} with id {self.Id}"
            )


class UpliftVanParticleSwarmResult(DStabilitySubStructure):
    FactorOfSafety: Optional[float] = None
    Id: Optional[str] = None
    LeftCenter: Optional[PersistablePoint] = None
    Points: Optional[List[Optional[PersistablePoint]]] = None
    RightCenter: Optional[PersistablePoint] = None
    Slices: Optional[List[Optional[PersistableSlice]]] = None
    TangentLine: Optional[float] = None

    @classmethod
    def structure_group(cls) -> str:
        return "results/upliftvanparticleswarm/"

    def get_slipcircle_output(self) -> UpliftVanSlipCircleResult:
        """Get condensed slipplane data"""
        try:
            return UpliftVanSlipCircleResult(
                x_left=self.LeftCenter.X,
                z_left=self.LeftCenter.Z,
                x_right=self.RightCenter.X,
                z_right=self.RightCenter.Z,
                z_tangent=self.TangentLine,
            )
        except (ValidationError, AttributeError):
            raise ValueError(
                f"Slipcircle not available for {self.__class__.__name__} with id {self.Id}"
            )


class UpliftVanReliabilityResult(DStabilitySubStructure):
    Converged: Optional[bool] = None
    FailureProbability: Optional[float] = None
    Id: Optional[str] = None
    LeftCenter: Optional[PersistablePoint] = None
    ReliabilityIndex: Optional[float] = None
    RightCenter: Optional[PersistablePoint] = None
    SoilContributions: Optional[List[Optional[PersistableSoilContribution]]] = None
    StageContributions: Optional[List[Optional[PersistableStageContribution]]] = None
    StateLinePointContributions: Optional[
        List[Optional[PersistableStateLinePointContribution]]
    ] = None
    StatePointContributions: Optional[
        List[Optional[PersistableStatePointContribution]]
    ] = None
    TangentLine: Optional[float] = None

    @classmethod
    def structure_group(cls) -> str:
        return "results/upliftvanreliability/"

    def get_slipcircle_output(self) -> UpliftVanSlipCircleResult:
        """Get condensed slipcircle data"""
        try:
            return UpliftVanSlipCircleResult(
                x_left=self.LeftCenter.X,
                z_left=self.LeftCenter.Z,
                x_right=self.RightCenter.X,
                z_right=self.RightCenter.Z,
                z_tangent=self.TangentLine,
            )
        except (ValidationError, AttributeError):
            raise ValueError(
                f"Slipcircle not available for {self.__class__.__name__} with id {self.Id}"
            )


class UpliftVanResult(DStabilitySubStructure):
    FactorOfSafety: Optional[float] = None
    Id: Optional[str] = None
    LeftCenter: Optional[PersistablePoint] = None
    Points: Optional[List[Optional[PersistablePoint]]] = None
    RightCenter: Optional[PersistablePoint] = None
    Slices: Optional[List[Optional[PersistableSlice]]] = None
    TangentLine: Optional[float] = None

    @classmethod
    def structure_group(cls) -> str:
        return "results/upliftvan/"

    def get_slipcircle_output(self) -> UpliftVanSlipCircleResult:
        """Get condensed slipcircle data"""
        try:
            return UpliftVanSlipCircleResult(
                x_left=self.LeftCenter.X,
                z_left=self.LeftCenter.Z,
                x_right=self.RightCenter.X,
                z_right=self.RightCenter.Z,
                z_tangent=self.TangentLine,
            )
        except (ValidationError, AttributeError):
            raise ValueError(
                f"Slipcircle not available for {self.__class__.__name__} with id {self.Id}"
            )


DStabilityResult = Union[
    UpliftVanResult,
    UpliftVanParticleSwarmResult,
    UpliftVanReliabilityResult,
    SpencerGeneticAlgorithmResult,
    SpencerReliabilityResult,
    SpencerResult,
    BishopBruteForceResult,
    BishopReliabilityResult,
    BishopResult,
]

###########################
# INPUT AND OUTPUT COMBINED
###########################


class DStabilityStructure(BaseModelStructure):
    """Highest level DStability class that should be parsed to and serialized from.

    The List[] items (one for each stage in the model) will be stored in a subfolder
    to multiple json files. Where the first (0) instance
    has no suffix, but the second one has (1 => _1) etc.

    also parses the outputs which are part of the json files
    """

    # input part
    waternets: List[Waternet] = [Waternet(Id="14")]  # waternets/waternet_x.json
    waternetcreatorsettings: List[WaternetCreatorSettings] = [
        WaternetCreatorSettings(Id="15")
    ]  # waternetcreatorsettings/waternetcreatorsettings_x.json
    states: List[State] = [State(Id="16")]  # states/states_x.json
    statecorrelations: List[StateCorrelation] = [
        StateCorrelation(Id="17")
    ]  # statecorrelations/statecorrelations_x.json
    stages: List[Stage] = [
        Stage(
            CalculationSettingsId="20",
            DecorationsId="12",
            GeometryId="11",
            Id="0",
            Label="Stage 1",
            LoadsId="18",
            Notes="Default stage by GEOLib",
            ReinforcementsId="19",
            SoilLayersId="13",
            StateId="16",
            StateCorrelationsId="17",
            WaternetCreatorSettingsId="15",
            WaternetId="14",
        )
    ]  # stages/stage_x.json
    soillayers: List[SoilLayerCollection] = [
        SoilLayerCollection(Id="13")
    ]  # soillayers/soillayers_x.json
    soilcorrelation: SoilCorrelation = SoilCorrelation()  # soilcorrelations.json
    soils: SoilCollection = SoilCollection()  # soils.json
    soilvisualizations: SoilVisualisation = SoilVisualisation()  # soilvisualizations.json
    reinforcements: List[Reinforcements] = [
        Reinforcements(Id="19")
    ]  # reinforcements/reinforcements_x.json
    projectinfo: ProjectInfo = ProjectInfo()  # projectinfo.json
    nailproperties: NailProperties = NailProperties()  # nailpropertiesforsoils.json
    loads: List[Loads] = [Loads(Id="18")]  # loads/loads_x.json
    decorations: List[Decorations] = [
        Decorations(Id="12")
    ]  # decorations/decorations_x.json
    calculationsettings: List[CalculationSettings] = [
        CalculationSettings(Id="20")
    ]  # calculationsettings/calculationsettings_x.json
    geometries: List[Geometry] = [Geometry(Id="11")]  # geometries/geometry_x.json

    # Output parts
    uplift_van_results: List[UpliftVanResult] = []
    uplift_van_particle_swarm_results: List[UpliftVanParticleSwarmResult] = []
    uplift_van_reliability_results: List[UpliftVanReliabilityResult] = []
    spencer_genetic_algorithm_results: List[SpencerGeneticAlgorithmResult] = []
    spencer_reliability_results: List[SpencerReliabilityResult] = []
    spencer_results: List[SpencerResult] = []
    bishop_bruteforce_results: List[BishopBruteForceResult] = []
    bishop_reliability_results: List[BishopReliabilityResult] = []
    bishop_results: List[BishopResult] = []

    @root_validator(skip_on_failure=True, allow_reuse=True)
    def ensure_validity_foreign_keys(cls, values):
        """TODO Include more fk relations, left for another issue."""
        for i, stage in enumerate(values.get("stages")):
            if stage.CalculationSettingsId != values.get("calculationsettings")[i].Id:
                raise ValueError("CalculationSettingsIds not linked!")
            if stage.DecorationsId != values.get("decorations")[i].Id:
                raise ValueError("DecorationsIds not linked!")
            if stage.GeometryId != values.get("geometries")[i].Id:
                raise ValueError("GeometryIds not linked!")
            if stage.LoadsId != values.get("loads")[i].Id:
                raise ValueError("LoadsIds not linked!")
            if stage.ReinforcementsId != values.get("reinforcements")[i].Id:
                raise ValueError("ReinforcementsIds not linked!")
            if stage.SoilLayersId != values.get("soillayers")[i].Id:
                raise ValueError("SoilLayersIds not linked!")
            if stage.StateId != values.get("states")[i].Id:
                raise ValueError("StateIds not linked!")
            if stage.StateCorrelationsId != values.get("statecorrelations")[i].Id:
                raise ValueError("StateCorrelationsIds not linked!")
            if (
                stage.WaternetCreatorSettingsId
                != values.get("waternetcreatorsettings")[i].Id
            ):
                raise ValueError("WaternetCreatorSettingsIds not linked!")
            if stage.WaternetId != values.get("waternets")[i].Id:
                raise ValueError("WaternetIds not linked!")
        return values

    @property
    def stage_specific_fields(self):
        return [
            "waternets",
            "waternetcreatorsettings",
            "states",
            "statecorrelations",
            "stages",
            "soillayers",
            "reinforcements",
            "loads",
            "decorations",
            "calculationsettings",
            "geometries",
        ]

    def get_stage_specific_fields(
        self, stage=0
    ) -> Generator[Tuple[str, DStabilitySubStructure], None, None]:
        """Yield stage specific fields for given stage."""
        for fieldname in self.stage_specific_fields:
            field = getattr(self, fieldname)
            if len(field) > stage:
                yield fieldname, field[stage]

    def renumber_fk_fields(self, instance, mapping: Dict, unique_id: int) -> int:
        """Replace id (foreign key) fields on instance based on a mapping and unique id."""
        fk = ForeignKeys()
        fkfields = fk.class_fields

        def get_correct_key(key, mapping):
            if key not in mapping:
                nonlocal unique_id
                mapping[key] = unique_id
                unique_id += 1
            return mapping[key]

        for fkfield in fkfields.get(instance.__class__.__name__, []):
            value = getattr(instance, fkfield)
            if isinstance(value, (list, set, tuple)):
                setattr(
                    instance,
                    fkfield,
                    [get_correct_key(x, mapping) for x in value],
                )
            if isinstance(value, (int, float, str)):
                setattr(instance, fkfield, get_correct_key(value, mapping))

        return unique_id

    def duplicate_stage(
        self, current_stage: int, label: str, notes: str, unique_start_id: int
    ):
        """Duplicates an existing stage.
        Copies the specific stage fields for a stage and renumbers all Ids,
        taking into account foreign keys by using the same renumbering.
        """

        old_to_new = {}
        for fieldname, stagefield in self.get_stage_specific_fields(current_stage):
            newstagefield = stagefield.copy(deep=True)

            # Renumber the upper class
            unique_start_id = self.renumber_fk_fields(
                newstagefield, old_to_new, unique_start_id
            )
            # Renumber all children
            for classinstance in children(newstagefield):
                unique_start_id = self.renumber_fk_fields(
                    classinstance, old_to_new, unique_start_id
                )

            # Update the stage with extra supplied fields
            if fieldname == "stages":
                newstagefield.Label = label
                newstagefield.Notes = notes

            getattr(self, fieldname).append(newstagefield)

        return len(self.stages) - 1, unique_start_id

    def add_default_stage(
        self, label: str, notes: str, unique_start_id=500
    ) -> Tuple[int, int]:
        """Add a new default (empty) stage to DStability."""
        self.waternets += [Waternet(Id=str(unique_start_id + 1))]
        self.waternetcreatorsettings += [
            WaternetCreatorSettings(Id=str(unique_start_id + 2))
        ]
        self.states += [State(Id=str(unique_start_id + 3))]
        self.statecorrelations += [StateCorrelation(Id=str(unique_start_id + 4))]
        self.soillayers += [SoilLayerCollection(Id=str(unique_start_id + 5))]
        self.soilcorrelation: SoilCorrelation = SoilCorrelation()
        self.reinforcements += [Reinforcements(Id=str(unique_start_id + 6))]
        self.loads += [Loads(Id=str(unique_start_id + 7))]
        self.decorations += [Decorations(Id=str(unique_start_id + 9))]
        self.calculationsettings += [CalculationSettings(Id=str(unique_start_id + 10))]
        self.geometries += [Geometry(Id=str(unique_start_id + 8))]
        self.stages += [
            Stage(
                CalculationSettingsId=str(unique_start_id + 10),
                DecorationsId=str(unique_start_id + 9),
                GeometryId=str(unique_start_id + 8),
                Id=str(unique_start_id),
                Label=label,
                LoadsId=str(unique_start_id + 7),
                Notes=notes,
                ReinforcementsId=str(unique_start_id + 6),
                SoilLayersId=str(unique_start_id + 5),
                StateId=str(unique_start_id + 3),
                StateCorrelationsId=str(unique_start_id + 4),
                WaternetCreatorSettingsId=str(unique_start_id + 2),
                WaternetId=str(unique_start_id + 1),
            )
        ]

        return len(self.stages) - 1, unique_start_id + 11

    def get_unique_id(self) -> int:
        """Return unique id that can be used in DStability.
        Finds all existing ids, takes the max and does +1.
        """

        fk = ForeignKeys()
        classfields = fk.class_fields

        ids = []
        for instance in children(self):
            for field in classfields.get(instance.__class__.__name__, []):
                value = getattr(instance, field)
                if isinstance(value, (list, set, tuple)):
                    ids.extend(value)
                if isinstance(value, (int, float, str)):
                    ids.append(value)

        new_id = max({int(id) for id in ids if id is not None}) + 1
        return new_id

    def validator(self):
        return DStabilityValidator(self)

    def has_stage(self, stage_id: int) -> bool:
        try:
            self.stages[stage_id]
            return True
        except IndexError:
            return False

    def has_result(self, stage_id: int) -> bool:
        if self.has_stage(stage_id):
            result_id = self.stages[stage_id].ResultId
            if result_id is None:
                return False
            else:
                return True
        return False

    def has_loads(self, stage_id: int) -> bool:
        if self.has_stage(stage_id):
            loads_id = self.stages[stage_id].LoadsId
            if loads_id is None:
                return False
            else:
                return True
        return False

    def has_soil_layers(self, stage_id: int) -> bool:
        if self.has_stage(stage_id):
            soil_layers_id = self.stages[stage_id].SoilLayersId
            if soil_layers_id is None:
                return False
            else:
                return True
        return False

    def has_soil_layer(self, stage_id: int, soil_layer_id: int) -> bool:
        if self.has_soil_layers(stage_id):
            for layer in self.soillayers[stage_id].SoilLayers:
                if str(soil_layer_id) == layer.LayerId:
                    return True
            return False
        return False

    def has_reinforcements(self, stage_id: int) -> bool:
        if self.has_stage(stage_id):
            reinforcements_id = self.stages[stage_id].ReinforcementsId
            if reinforcements_id is None:
                return False
            else:
                return True
        return False

    def get_result_substructure(
        self, analysis_type: AnalysisTypeEnum, calculation_type: CalculationTypeEnum
    ) -> List[DStabilityResult]:

        result_types_mapping = {
            AnalysisTypeEnum.UPLIFT_VAN: {
                "non_probabilistic": self.uplift_van_results,
                "probabilistic": self.uplift_van_reliability_results,
            },
            AnalysisTypeEnum.UPLIFT_VAN_PARTICLE_SWARM: {
                "non_probabilistic": self.uplift_van_particle_swarm_results,
                "probabilistic": self.uplift_van_reliability_results,
            },
            AnalysisTypeEnum.SPENCER_GENETIC: {
                "non_probabilistic": self.spencer_genetic_algorithm_results,
                "probabilistic": self.spencer_reliability_results,
            },
            AnalysisTypeEnum.SPENCER: {
                "non_probabilistic": self.spencer_results,
                "probabilistic": self.spencer_reliability_results,
            },
            AnalysisTypeEnum.BISHOP_BRUTE_FORCE: {
                "non_probabilistic": self.bishop_bruteforce_results,
                "probabilistic": self.bishop_reliability_results,
            },
            AnalysisTypeEnum.BISHOP: {
                "non_probabilistic": self.bishop_results,
                "probabilistic": self.bishop_reliability_results,
            },
        }

        if calculation_type == CalculationTypeEnum.PROBABILISTIC:
            return result_types_mapping[analysis_type]["probabilistic"]

        return result_types_mapping[analysis_type]["non_probabilistic"]


class ForeignKeys(DStabilityBaseModelStructure):
    """A dataclass that store the connections between the
    various unique Ids used in DStability. These can be seen
    as (implicit) foreign keys.
    """

    mapping: Dict[str, Tuple[str, ...]] = {
        "Waternet.Id": ("Stage.WaternetId",),
        "PersistableHeadLine.Id": (
            "PersistableReferenceLine.BottomHeadLineId",
            "PersistableReferenceLine.TopHeadLineId",
        ),
        "PersistableReferenceLine.Id": ("Waternet.PhreaticLineId",),
        "PersistableLayer.Id": (
            "PersistableStatePoint.LayerId",
            "PersistableSoilLayer.LayerId",
            "PersistableConsolidation.LayerId",
            "PersistableLayerLoad.LayerId",
            "PersistableBerm.AddedLayerId",
            "WaternetCreatorSettings.AquiferInsideAquitardLayerId",
            "WaternetCreatorSettings.AquiferLayerId",
        ),
        # Soil commented out for now, isn't used in stages
        # "PersistableSoil.Id": (
        #     "PersistableSoilVisualization.SoilId",
        #     "PersistableSoilLayer.SoilId",
        #     "PersistableSoilCorrelation.CorrelatedSoilIds",
        #     "PersistableNailPropertiesForSoil.SoilId",
        #     "PersistableSoilContribution.SoilId"
        # ),
        "CalculationSettings.Id": ("Stage.CalculationSettingsId",),
        "Decorations.Id": ("Stage.DecorationsId",),
        "Geometry.Id": ("Stage.GeometryId",),
        "Loads.Id": ("Stage.LoadsId",),
        "Reinforcements.Id": ("Stage.ReinforcementsId",),
        "Result.Id": ("Stage.ResultId",),
        "SoilLayerCollection.Id": ("Stage.SoilLayersId",),
        "StateCorrelation.Id": ("Stage.StateCorrelationsId",),
        "State.Id": ("Stage.StateId",),
        "WaternetCreatorSettings.Id": ("Stage.WaternetCreatorSettingsId",),
        "Stage.Id": ("PersistableStageContribution.StageId",),
        "PersistableStateLinePoint.Id": (
            "PersistableStateCorrelation.CorrelatedStateIds",
            "PersistableStateLinePointContribution.StateLinePointId",
        ),
        "PersistableStatePoint.Id": ("PersistableStatePointContribution.StatePointId",),
    }

    @property
    def class_fields(self) -> Dict[str, List[str]]:
        """Return a mapping in the form:
        classname: [fields]
        """
        id_keys = chain(*((k, *v) for k, v in self.mapping.items()))
        class_fields = defaultdict(list)
        for id_key in id_keys:
            classname, fieldname = id_key.split(".")
            class_fields[classname].append(fieldname)
        return class_fields



from typing import List, Tuple, Type

from pydantic import FilePath

from geolib.models.dseries_parser import DSerieParser, DSeriesStructure
from geolib.models.parsers import BaseParserProvider

from .internal import DSettlementOutputStructure, DSettlementStructure


class DSettlementInputParser(DSerieParser):
    """DSettlement parser of input files."""

    @property
    def suffix_list(self) -> List[str]:
        return [".sli"]

    @property
    def dserie_structure(self) -> Type[DSettlementStructure]:
        return DSettlementStructure


class DSettlementOutputParser(DSerieParser):
    """DSettlement parser of input files."""

    @property
    def suffix_list(self) -> List[str]:
        return [".sld"]

    @property
    def dserie_structure(self) -> Type[DSettlementOutputStructure]:
        return DSettlementOutputStructure


class DSettlementParserProvider(BaseParserProvider):

    __input_parsers = None
    __output_parsers = None

    @property
    def input_parsers(self) -> Tuple[DSettlementInputParser]:
        if not self.__input_parsers:
            self.__input_parsers = (DSettlementInputParser(),)
        return self.__input_parsers

    @property
    def output_parsers(self) -> Tuple[DSettlementOutputParser]:
        if not self.__output_parsers:
            self.__output_parsers = (DSettlementOutputParser(),)
        return self.__output_parsers

    @property
    def parser_name(self) -> str:
        return "DSettlement"



from typing import List, Optional

from pydantic import confloat, conlist, constr, validator

from geolib.geometry import Point
from geolib.models import BaseDataClass

from .internal import Surface as InternalSurface
from .settings import DistributionType


class Surface(BaseDataClass):
    """Surface.

    Args:
        name: Name of the surface.
        points: Surface points. Points (X, Z) must be defined from the sheet outwards regarding x-coordinate.
        distribution_type: Distribution type.
        std: Standard deviation of the distribution type.
    """

    name: constr(min_length=1, max_length=50)
    points: conlist(Point, min_items=1)
    distribution_type: Optional[DistributionType] = None
    std: Optional[confloat(ge=0.0)] = None

    @validator("points")
    def points_must_be_increasing_and_greater_or_equal_to_zero(cls, v):
        x_coords = [p.x for p in v]
        if min(x_coords) < 0:
            raise ValueError(
                f"All x-coordinates must be greater than or equal to 0, found {min(x_coords)}"
            )
        if x_coords[0] != 0:
            raise ValueError(
                f"X-coordinate first point should be zero, received {x_coords[0]}"
            )
        if x_coords != sorted(x_coords):
            raise ValueError("x-coordinates must be strictly increasing")
        return v

    def to_internal(self) -> InternalSurface:

        kwargs = self.dict(exclude_none=True, exclude={"points"})
        kwargs["points"] = [
            {"Nr": i, "X-coord": p.x, "Value": p.z}
            for i, p in enumerate(self.points, start=1)
        ]
        return InternalSurface(**kwargs)



from __future__ import annotations

import inspect
import logging
import re
import shlex
from abc import abstractclassmethod, abstractmethod
from itertools import groupby
from math import isfinite
from typing import (
    Dict,
    Iterable,
    List,
    Tuple,
    Type,
    Union,
    _GenericAlias,
    _SpecialForm,
    get_type_hints,
)

from pydantic import FilePath

from geolib.errors import ParserError
from geolib.models import BaseDataClass as DataClass
from geolib.models.base_model_structure import BaseModelStructure

from .parsers import BaseParser
from .utils import (
    get_args,
    get_filtered_type_hints,
    get_required_class_field,
    is_list,
    is_union,
    unpack_if_union,
)

logger = logging.getLogger(__name__)


class DSeriesStructure(BaseModelStructure):
    def __init__(self, *args, **kwargs):
        """The base class for all DSerie structures.

        It's parent is a BaseModel from Pydantic, which expects
        all its fields named as kwargs.

        Here we check, using the type annotations stored for each
        field, whether we can parse/convert a field, passed as string,
        to a more concrete implementation.

        When we have a class:
        ```
        Test(DSeriesStructure):
            field_a: Union[OtherClass, str]
        ```
        and we receive field_a as a `str`, we check the attributes
        of `OtherClass` to see if it (also) has a `is_parseable` method.
        If so, we convert it from `str` to `OtherClass` using its
        `parse_text` method.

        If `OtherClass` is a child of `DSeriesStructure`, this process
        is recursive.

        Note that the heuristics of checking type hints such as
        `Optional` and `List` or `Union` are still archaic.
        In later versions of Python this is improved.
        """
        # Remove fields that are None so defaults will be used
        kwargs = {field: value for field, value in kwargs.items() if value is not None}

        if len(kwargs) > len(get_type_hints(self)):
            a = set(kwargs.keys())
            b = set(get_type_hints(self).keys())
            raise ValueError(
                f"""Got more fields than defined on model {self.__class__.__name__}:
                parser has {a.difference(b)} fields and
                model has {b.difference(a)} fields not set.
                """
            )

        for field, fieldtype in get_type_hints(self).items():

            # If the body is a string, we should check
            # whether we can parse it further.
            if field in kwargs and isinstance(kwargs[field], str):
                body = kwargs[field]

                # Optional, Union, etc.
                if is_union(fieldtype):
                    fieldtype, *_ = get_args(fieldtype)

                if is_list(fieldtype):
                    fieldtype, *_ = get_args(fieldtype)

                if hasattr(fieldtype, "is_parseable") and fieldtype.is_parseable():
                    logger.debug(f"Changed {field} to {fieldtype}")
                    kwargs[field] = fieldtype.parse_text(body)
                else:
                    if DataClass in fieldtype.__mro__:
                        logger.debug(f"Can't parse {fieldtype} for {field} yet")

            # If the body is a List[string], we should check
            # whether we can parse it further.
            elif (
                field in kwargs
                and isinstance(kwargs[field], list)
                and len(kwargs[field]) > 0
                and isinstance(kwargs[field][0], str)
            ):
                body = kwargs[field]

                # We need to decompose a possible Union
                if is_union(fieldtype):
                    fieldtype, *_ = get_args(fieldtype)

                # And then the list
                if is_list(fieldtype):
                    fieldtype, *_ = get_args(fieldtype)

                    # Check whether we can parse and if so, parse the strings
                    if hasattr(fieldtype, "is_parseable") and fieldtype.is_parseable():
                        logger.debug(f"Changed {field} to multiple {fieldtype}")
                        kwargs[field] = [fieldtype.parse_text(item) for item in body]
                    else:
                        if DataClass in fieldtype.__mro__:
                            logger.debug(f"Can't parse {fieldtype} for {field} yet")

                else:
                    # Check whether we can parse and if so, parse the strings
                    if hasattr(fieldtype, "is_parseable") and fieldtype.is_parseable():
                        logger.debug(
                            f"Changed {field} to single {fieldtype} for list of str."
                        )
                        kwargs[field] = fieldtype.parse_text(body)
                    else:
                        if DataClass in fieldtype.__mro__:
                            logger.debug(f"Can't parse {fieldtype} for {field} yet")

            else:
                # Ignore other fields
                continue

        super().__init__(**kwargs)

    @staticmethod
    def is_parseable() -> bool:
        return True

    @classmethod
    def parse_text(cls, data: str) -> DSeriesStructure:
        """Creates a DSeriesStructure using the concrete parser.

        Arguments:
            data {str} -- Data to parse.
        """
        parsed_structure = DSerieParser.parse_group_as_dict(data)
        return cls(**parsed_structure)

    @classmethod
    def get_structure_required_fields(cls) -> List[Tuple[str, Type]]:
        """Allows concrete classes to filter out fields that would normally be
        parsed.

        Returns:
            List[Tuple[str, Type]]: List of Tuples[FieldName, FieldType]
        """
        return get_filtered_type_hints(cls)

    def dict(_, *args, **kwargs):
        data = super().dict(*args, **kwargs)
        for __, value in data.items():
            if isinstance(value, float) and not isfinite(value):
                raise ValueError(
                    "Only finite values are supported, don't use nan, -inf or inf."
                )
        return data


class DSerieListGroupNextStructure(DSeriesStructure):
    """Creates a DSeriesStructure that can parse fields like this:

    [NUMBER OF STRUCTURES]
    2
    -------
    [STRUCTURE VALUE]
    42
    [END STRUCTURE VALUE]
    [NEXT NUMBER OF STRUCTURES]
    [STRUCTURE VALUE]
    24
    [END STRUCTURE VALUE]
    [END NUMBER OF STRUCTURES]

    into:

    wrapper = [
        "2\n-------\n[STRUCTURE VALUE...",
        "[STRUCTURE VALUE...",
    ]

    Note the "NEXT NUMBER OF STRUCTURES" element that makes this different
    from a normal ListGroup structure. The extra strings at the beginning
    of the first element should be ignored by the child parser.
    """

    @abstractclassmethod
    def group_delimiter(cls) -> str:
        raise NotImplementedError("Implement in derived classes.")

    @classmethod
    def parse_text(cls, text: str):
        # Strip outer wrapper
        body = list(DSerieParser.parse_group(text))[0][1]

        num_elements = int(body.strip().split("\n")[0])
        elements = body.split(cls.group_delimiter())
        # Check whether we find the number of elements
        # as stated in the header.
        if num_elements == 0:
            # Even if there are zero elements, the split() method will still return a list of length 1
            # "a".split("b") -> ["a"]
            if not len(elements) - 1 == num_elements:
                raise ParserError(f"Error in parsing in number of elements for {cls}")
            elements = []
        else:
            if not len(elements) == num_elements:
                raise ParserError(f"Error in parsing in number of elements for {cls}")

        collection_property_name = list(cls.__fields__.items())[0][0]
        return cls(**{collection_property_name: elements})


class DSeriesTableStructure(DSeriesStructure):
    @classmethod
    def parse_text(cls, text: str):
        """Creates a DSeriesStructure that can parse fields like

        [GROUP]
        [COLUMN INDICATION]
        A
        B
        C
        [END OF COLUMN INDICATION]
        [GROUP DATA]
        3
        1 1 1
        2 2 2
        2 3 4
        [END OF GROUP DATA]
        [END OF GROUP]

        returning:
        [
            {"A": 1, "B": 1, "C": 1},
            {"A": 1, "B": 1, "C": 1},
            {"A": 2, "B": 3, "C": 4}
        ]

        The group name can differ and is not parsed,
        but the classname is used as fieldname.

        Arguments:
            data {str} -- Data to parse.
        """
        groups = DSerieParser.parse_group_as_dict(text)
        if not len(groups) == 2:
            raise ParserError(f"Error parsing for {cls}, expected only 2 elements.")
        for groupname, text in groups.items():
            if groupname == "column_indication":
                columns = [make_key(line) for line in text.split("\n") if line != ""]
            else:
                lines = [
                    split_line_elements(line) for line in text.split("\n") if line != ""
                ]
                cls.validate_number_of_rows(lines)
        lines = [dict(zip(columns, parts)) for parts in lines]

        d = {cls.__name__.lower(): lines}
        return cls(**d)

    @classmethod
    def validate_number_of_rows(cls, lines: List[List[str]]):
        """Validates whether the number of lines matched the expected rows to be read.

        Args:
            lines (List[List[str]]): Lines representing rows of data.
        Raises:
            ParserError: When the expectations are not met.
        """
        count = int(lines.pop(0)[0])
        if not count == len(lines):
            raise ParserError(
                f"Error parsing for {cls}, header indicates {count} lines, while there are {len(lines)} lines."
            )


class DSeriesWrappedTableStructure(DSeriesTableStructure):
    @classmethod
    def validate_number_of_rows(cls, lines):
        """These sort of tables do not contain information of the number of rows
        contained in the file. Therefore they are not validated.

        Args:
            lines (List[str]): Parsed rows.
        """
        pass

    @classmethod
    def parse_text(cls, text):
        unwrapped_text = DSerieParser.parse_group_as_dict(text)
        return super().parse_text(list(unwrapped_text.values())[0])


class DSerieOldTableStructure(DSeriesStructure):
    @classmethod
    def parse_text(cls, text: str):
        """Creates a DSeriesStructure that can parse fields like

        [GROUP]
        [COLUMN INDICATION]
        A
        B
        C
        [END OF COLUMN INDICATION]
        [DATA COUNT]
        3
        [END OF DATA COUNT]
        [GROUP DATA]
        1 1 1
        2 2 2
        2 3 4
        [END OF GROUP DATA]
        [END OF GROUP]

        returning:
        [
            {"A": 1, "B": 1, "C": 1},
            {"A": 1, "B": 1, "C": 1},
            {"A": 2, "B": 3, "C": 4}
        ]

        The group name can differ and is not parsed,
        but the classname is used as fieldname.

        Arguments:
            data {str} -- Data to parse.
        """
        groups = DSerieParser.parse_group_as_dict(text)
        if not len(groups) == 3:
            raise ParserError(f"Error parsing for {cls}, expected only 3 elements.")
        for groupname, text in groups.items():
            if groupname == "column_indication":
                columns = [make_key(line) for line in text.split("\n") if line != ""]
            elif groupname == "data_count":
                lines = [
                    split_line_elements(line) for line in text.split("\n") if line != ""
                ]
                count = int(lines[0][0])
            else:
                lines = [
                    split_line_elements(line) for line in text.split("\n") if line != ""
                ]
                if not count == len(lines):
                    raise ParserError(
                        f"Error parsing for {cls}, header indicates {count} lines, while there are {len(lines)} lines."
                    )

        lines = [dict(zip(columns, parts)) for parts in lines]
        d = {cls.__name__.lower(): lines}
        return cls(**d)


class DSerieListStructure(DSeriesStructure):
    @classmethod
    def parse_text(cls, text: str):
        """Creates a DSeriesStructure that can parse fields like

        [GROUP]
        3
        1
        1
        1
        [END OF GROUP]

        returning:
        [1,1,1]

        Arguments:
            data {str} -- Data to parse.
        """
        lines = [
            strip_line_first_element(line) for line in text.split("\n") if line != ""
        ]
        nrow = int(lines.pop(0))
        if not nrow == len(lines):
            raise ParserError(
                f"Error parsing for {cls}, header indicates {nrow} lines, while there are {len(lines)} lines."
            )
        d = {cls.__name__.lower(): lines}
        return cls(**d)

    @classmethod
    def parse_text_lines(cls, lines: List[List[str]]):
        d = {cls.__name__.lower(): [line[0] for line in lines]}
        return cls(**d), 1


class DSerieMatrixStructure(DSeriesStructure):
    @classmethod
    def parse_text(cls, text: str):
        """Creates a DSeriesStructure that can parse fields like

        [GROUP]
        3
        1 2
        1 2
        1 2
        [END OF GROUP]

        returning:

        [[1,2], [1,2], [1,2]]

        Arguments:
            data {str} -- Data to parse.
        """
        lines = [split_line_elements(line) for line in text.split("\n") if line != ""]
        nrow = int(lines.pop(0)[0])
        ncol = int(lines.pop(0)[0]) + 1
        if not nrow == len(lines):
            raise ParserError(
                f"Error parsing for {cls}, header indicates {nrow} lines, while there are {len(lines)} lines."
            )
        if not ncol == len(lines[0]):
            raise ParserError(
                f"Error parsing for {cls}, header indicates {ncol} columns, while there are {len(lines[0])} columns."
            )
        d = {cls.__name__.lower(): lines}
        return cls(**d)


class DSeriesInlineProperties(DSeriesStructure):
    """Generates a structure where the properties are assumed to be
    in the same order as define in the concrete implementation.
    Hence, only the read value field is relevant here.
    Use when properties are in order.
    Example:
        Given class my_structure(DSeriesInlineProperties):
            property_one: int
        and text, text_to_parse:
            [InlineStructure]
            dummy_text: 42 some unit/and scale
            [END OF InlineStructure]
        When ms = my_structure.parse_text(text)
        Then:
            parsed_structure.property_one = 42

    Returns:
        DSeriesStructure: Parsed structure.
    """

    @classmethod
    def header_lines(cls) -> int:
        """Returns the amount of header lines expected for the text.
        Concrete classes can override this method.

        Returns:
            int: Number of header lines to skip.
        """
        return 0

    @classmethod
    def get_property_key_value(cls, text: str, expected_property: str) -> Tuple[str, str]:
        """Gets the property key and value for a given text line.
        It allows concrete classes to override it and either use the expected property
        name or another value.

        Args:
            text (str): Text line to parse.
            expected_property (str): The expected property key if parsing the structure properties in order.

        Returns:
            Tuple[str, str]: Property name and property value text.
        """
        # If there's no key it's because this is actually a property,
        # the real key and values are stored in the second item of the tuple.
        if text:
            # If the value is not empty, extract its value (assume it's on the first position)
            # Note: If that is the case, you should be using other structure such as:
            # DSeriesRepeatedGroupsWithInlineMappedProperties
            _, value = get_line_property_key_value(text, reversed_key=False)
            return expected_property, value
        return expected_property, text

    @classmethod
    def get_properties_in_text(cls, text: str):
        required_properties_names = [
            structure_name for structure_name, _ in cls.get_structure_required_fields()
        ]
        header_lines = cls.header_lines()
        for idx, (key, value) in enumerate(
            DSerieParser.parse_group(text, loose_properties=True)
        ):
            if idx < header_lines:
                continue
            if not key:
                # Only loose properties would get through here.
                # So a grouped property enclosed like [PROPERTY][END OF PROPERTY].
                key, value = cls.get_property_key_value(
                    value, required_properties_names[idx - header_lines]
                )
            yield key, value

    @classmethod
    def parse_text(cls, text: str) -> DSeriesStructure:
        """Parses a text containing a list of inline properties which are
        mapped to the properties of the concrete class.
        Example:
            Given:
                [group]
                4.2 = property_one
                2.4 : property_two
                [property_in_group]
                42
                [end of property_in_group]
                [end of group]
            Returns:
                {
                    property_one: "4.2",
                    property_two: "2.4",
                    property_in_group: "42",
                }
        Args:
            text (str): Inline properties list.

        Returns:
            DSeriesStructure: Parsed structure.
        """
        return cls(**{key: value for key, value in cls.get_properties_in_text(text)})


class DSeriesInlineMappedProperties(DSeriesInlineProperties):
    """Specialization of DSeriesInlineProperties, use this class when the
    properties in the text are the same as in the concrete class definition:
    Example:
        Given class my_structure(DSeriesInlineMappedProeprties):
            property_dummy_int: int
            property_dummy_str: str
        and text, text_to_parse:
            [STRUCTURE]
            property dummy int = 42 some unit/scale
            [PROPERTY DUMMY STR]
            It also works with mixed groups.
            [END OF PROPERTY DUMMY STR]
            [END OF STRUCTURE]
        When ms = my_structure.parse_text(text_to_parse)
        Then:
            ms.property_dummy_int = 42
            ms.proeprty_dummy_str = "It also works with mixed groups."
    """

    @classmethod
    def get_property_key_value(cls, text: str, expected_property: str) -> Tuple[str, str]:
        """Gets both property key and value from the text line.

        Args:
            text (str): Text containing key and value.
            expected_property (str): The expected property if there was no key.

        Returns:
            Tuple[str, str]: Parsed key and value.
        """
        key = expected_property
        unformatted_key, value = get_line_property_key_value(text, reversed_key=False)
        if unformatted_key:
            key = make_key(unformatted_key)
        return key, value


class DSerieVersion(DSeriesInlineMappedProperties):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for k, v in self.dict().items():
            if self.__fields__.get(k).get_default() != v:
                logger.warning(
                    """The version of the input file is unsupported.
                Check the documentation on how to prevent this warning in the future."""
                )
                break


class DSeriesInlineReversedProperties(DSeriesInlineProperties):
    @classmethod
    def get_property_key_value(cls, text: str, expected_property: str) -> Tuple[str, str]:
        """Returns the value content for a line of format:
        value : key || value = key

        Args:
            text (str): Text line with value-key

        Returns:
            str: Filtered value.
        """
        if text:
            # If the value is not empty, extract its value (assume it's on the first position)
            _, value = get_line_property_key_value(text, reversed_key=True)
            return expected_property, value
        return expected_property, text


class DSeriesUnmappedNameProperties(DSeriesInlineMappedProperties):
    """Specialization of DSeriesInlineMappedProperties, use this class when the
    properties in the text are the same as in the concrete class definition except
    for "name", which has not been declared as a key in the input text:
    Example:
        Given class my_structure(DSeriesInlineMappedProeprties):
            name: str
            property_dummy_int: 42
        and text, text_to_parse:
            [STRUCTURE]
            This is a name
            property dummy int = 42 some unit/scale
            [END OF STRUCTURE]
        When ms = my_structure.parse_text(text_to_parse)
        Then:
            ms.name: "This is a name"
            ms.property_dummy_int = 42
    """

    @classmethod
    def get_property_key_value(cls, text: str, expected_property: str) -> Tuple[str, str]:
        """Gets both property key and value from the text line.

        Args:
            text (str): Text containing key and value.
            expected_property (str): The expected property if there was no key.

        Returns:
            Tuple[str, str]: Parsed key and value.
        """

        if expected_property == "name":
            return expected_property, text.strip()
        return super().get_property_key_value(text, expected_property)


class DSeriesStructureCollection(DSeriesStructure):
    @classmethod
    def parse_text(cls, text: str) -> DSeriesStructure:
        """Parses a text containing a collection of DSeriesStructure.
        Example:
            Given:
                [STRUCTURE COLLECTION]
                    2 = Number of items
                    [STRUCTURE]
                    Property 1 = 1
                    Property 2 = 2
                    [END OF STRUCTURE]
                    [STRUCTURE]
                    Property 1 = 3
                    Property 2 = 4
                    [END OF STRUCTURE]
                [END OF STRUCTURE]
            Returns:
                [
                    structure({property_1: 1, property_2: 2}),
                    structure({property_1: 3, property_2: 3}),
                ]

        Args:
            text (str): Text containing collection to parse.

        Raises:
            ValueError: When number of read structures does not match declared expectation.
            ValueError: When number of concrete class properties is not equal to 1.

        Returns:
            DSeriesStructure: Parsed structure.
        """
        fields = cls.get_structure_required_fields()
        if len(fields) != 1:
            raise ValueError(
                f"This type of collection is only meant to have one field but {len(fields)} were defined."
            )

        parsed_groups = [
            text_value
            for _, text_value in DSerieParser.parse_group(text, loose_properties=True)
        ]
        number_of_groups = get_line_property_value(
            parsed_groups.pop(0), reversed_key=True
        )

        if len(parsed_groups) != int(number_of_groups):
            raise ValueError(
                f"Expected {number_of_groups} groups, but parsed {len(parsed_groups)}."
            )

        parsed_collection = {fields[0][0]: parsed_groups}
        return cls(**parsed_collection)


class DSeriesRepeatedGroupedProperties(DSeriesStructure):
    @classmethod
    def parse_text(cls, text: str) -> DSeriesStructure:
        """Parses a structure which fields are all encapsulated in subgroups,
        some of them might appear repeated and they have to be grouped in lists.
        Example:
            Given input:
                [PROPERTY ONE]
                    42
                    2.4
                [END OF PROPERTY ONE]
                [PROPERTY ONE]
                    24
                    4.2
                [END OF PROPERTY ONE]
            Returns:
                {'property_one': ['42\n2.4\n', '24\n4.2\n']}

        Args:
            text (str): Text to parse.

        Returns:
            DSeriesStructure: Parsed structrure.
        """
        generated_dict = {}
        no_key_group_values = []
        for group_key, group_values in groupby(
            list(DSerieParser.parse_group(text, loose_properties=True)),
            lambda key_value: key_value[0],
        ):
            filtered_values = [value for key, value in group_values]
            if not group_key:
                no_key_group_values.extend(filtered_values)
                continue
            if cls.group_value_is_list(group_key):
                if group_key in generated_dict:
                    generated_dict[group_key].extend(filtered_values)
                else:
                    generated_dict[group_key] = filtered_values
            else:
                generated_dict[group_key] = filtered_values[0]

        generated_dict.update(cls.get_inline_properties(no_key_group_values))
        return cls(**cls.get_validated_mappings(generated_dict))

    @classmethod
    def group_value_is_list(cls, group_key: str) -> bool:
        """Returns whether a group key should be parsed as list or flat value.
        Allows inherited classes to override its behavior.

        Args:
            group_key (str): Property name.

        Returns:
            bool: Whether the key represents a list of values or not.
        """
        return group_key in cls.get_list_field_names()

    @classmethod
    def get_list_field_names(cls) -> List[str]:
        """Returns the field names which are typed as list.
        Allows for extension on concrete classes.

        Returns:
            List[str]: List of field names.
        """
        return [
            field_name
            for field_name, field_type in get_type_hints(cls).items()
            if is_list(field_type)
        ]

    @classmethod
    def get_inline_properties(cls, inline_properties: List[str]) -> Dict[str, str]:
        """Processes all unmapped properties and returns them in a dictionary.
        This method can be replaced in concrete implementations of this class.
        Args:
            inline_properties (List[str]): List of property values which key was not found.
        Returns:
            Dict[str, str]: Returns a dictionary of inline properties.
        """
        logger.debug(
            "The following properties were not mapped because no key was found for them:"
            + "\n".join(inline_properties)
        )
        # This class should not return unmapped properties.
        return {}

    @classmethod
    def get_validated_mappings(cls, generated_dict: Dict[str, str]) -> Dict[str, str]:
        """Validates the input dictionary mapping within this class properties.
        Allows for extension on lower classes.

        Args:
            generated_dict (Dict[str, str]): [description]

        Returns:
            Dict[str, str]: Validated dictionary.
        """
        return generated_dict


class DSeriesRepeatedGroupsWithInlineMappedProperties(DSeriesRepeatedGroupedProperties):
    @classmethod
    def get_inline_properties(cls, inline_properties: List[str]) -> Dict[str, str]:
        """Processes all unmapped properties and returns them in a dictionary.
        This method can be replaced in concrete implementations of this class.
        Follows a philosophy of FIFO for repeated keys.
        Args:
            inline_properties (List[str]): List of property values which key was not found.
        Returns:
            Dict[str, str]: Resulting inline mapped properties.
        """
        result_dict: Dict[str, str] = {}
        for unmapped in inline_properties:
            unformatted_key, value = get_line_property_key_value(
                unmapped, reversed_key=False
            )
            key = make_key(unformatted_key)

            if not key:
                # If no value was found then we just need the stripped remaining.
                value = unmapped.strip()

            if key in result_dict:
                logger.debug(
                    f"Key {key} already mapped, value {value} will not be mapped."
                )
                continue
            result_dict[key] = value
        return result_dict


# region TODO: Deprecate.


class DSeriesTreeStructure(DSeriesStructure):
    # TODO Deprecate.
    # This class is overdoing logic that is later on being handled by its parent (DSeriesStructure).
    # It should therefore only be responsible for generating a dictionary of field name - string values,
    # but not object creation.
    @classmethod
    def parse_text(cls, text: str) -> DSeriesStructure:
        """Parses a text containing listed properties.
        Example:
        1 - structure id
        2 - number of values for property X:
            4 2

        Args:
            text (str): text to parse into properties.

        Returns:
            DSeriesStructure: Structure with parsed properties.
        """
        parsed_structure, _ = cls.parse_text_lines(
            [split_line_elements(line) for line in text.split("\n") if line != ""]
        )
        return parsed_structure

    @classmethod
    def parse_text_lines(
        cls, text_lines: List[List[str]]
    ) -> Tuple[DSeriesStructure, int]:
        """Parses a list of strings into the properties of a structure.
        Example:
            [   ["1", "Property value"],
                ["2", "Values in Property"],
                ["4", "2"]]
        Args:
            text_lines (List[List[str]]): List of properties as strings.

        Raises:
            ValueError: When the number of values for a property does not match its definition.
            ValueError: When the number of properties does not match the text definition.

        Returns:
            DSeriesStructure: Structure with parsed properties.
            int: Index of last line being read as part of current structure.
        """

        def get_list_values(
            struct_idx: int, field_name: str, text_lines: list
        ) -> Tuple[DSeriesStructure, int]:
            """Auxiliar method to either parse as a DSeriesStructure or as a list of
            any different object.

            Args:
                struct_idx (int): Type index in the class definition.
                field_name (str): Name of the field.
                text_lines (list): Lines containing values to parse.

            Returns:
                Tuple[DSeriesStructure, int]: Parsed structure and lines read.
            """
            list_type = get_field_collection_type(cls, struct_idx)
            if issubclass(list_type, DSeriesTreeStructureCollection):
                return list_type.parse_text_lines(text_lines)
            else:
                return read_property_as_list(field_name, list_type, text_lines)

        properties = {}
        required_fields = cls.get_structure_required_fields()

        if len(text_lines) < len(required_fields):
            raise ValueError(
                f"There should be at least {len(required_fields)}"
                + f" fields to correctly initalize object {cls}"
            )
        lines_read = 0
        for struct_idx, (field_name, field) in enumerate(required_fields):
            if lines_read >= len(text_lines):
                raise ValueError(f"Expected text line property for {field_name}.")
            iteration_lines = 1
            field = unpack_if_union(field)
            # if the current property is a list, then extract the next line values.
            if (
                is_list(field)
                or issubclass(field, list)
                or is_structure_collection(field)
            ):
                lines_to_parse = cls.get_next_property_text_lines(text_lines[lines_read:])
                parsed_tuple = get_list_values(struct_idx, field_name, lines_to_parse)
                (
                    properties[field_name],
                    iteration_lines,
                ) = cls.get_tree_structure_read_lines(parsed_tuple)
            elif field == str:  # for names in structures
                properties[field_name] = " ".join(text_lines[lines_read]).strip()
            else:
                properties[field_name] = get_line_property_value(
                    text_lines[lines_read][0], reversed_key=True
                )
            lines_read += iteration_lines
            if len(properties) == len(required_fields):
                break
        return cls(**properties), lines_read

    @classmethod
    def get_next_property_text_lines(cls, text_lines: list) -> list:
        """Allows inherited classes to override the text lines that are fed into other
        structures contained in a list property.

        Args:
            text_lines (list): List containing other lists or flat text to be parsed.

        Returns:
            list: Filtered list of text to be parsed into a property.
        """
        return text_lines

    @classmethod
    def get_tree_structure_read_lines(
        cls, parsed_tuple: Tuple[DSeriesStructure, int]
    ) -> Tuple[DSeriesStructure, int]:
        """Allows inherited classes to override how many lines have been actually been read
        for a given parsed structure based on the cls being instantiated.

        Args:
            parsed_tuple (Tuple[DSeriesStructure, int]): Tuple of parsed structure and read lines for it.

        Returns:
            Tuple[DSeriesStructure, int]: Resulting tuple with real number of parsed lines.
        """
        return parsed_tuple


class DSeriesTreeStructurePropertiesInGroups(DSeriesTreeStructure):
    # TODO Deprecate.
    # This class is overdoing logic that is later on being handled by its parent (DSeriesStructure).
    # It should therefore only be responsible for generating a dictionary of field name - string values,
    # but not object creation.
    @classmethod
    def parse_text(cls, text: str):
        """Creates a DSeriesStructure which properties are divided in subgroups.
        Example

        [PROPERTY_1]
        42
        [END OF PROPERTY_1]
        [PROPERTY_2]
        24, 42
        [END OF PROPERTY_2]
        returns:
        DSeriesTreeStructurePropertiesInGroups({
            property_1: 42,
            property_2: [24, 42]
        })

        Arguments:
            data {str} -- Data to parse.
        """
        tuple_lines = [[value, key] for key, value in DSerieParser.parse_group(text)]
        return cls.parse_text_lines(tuple_lines)[0]

    @classmethod
    def get_next_property_text_lines(cls, text_lines: list) -> list:
        """Retrieves the first element of a list of Tuples[PropertyTextValue, PropertyName]
        and splits it into further parseable lines.

        Args:
            text_lines (list): list of Tuples[PropertyTextValue, PropertyName].

        Returns:
            list: List of lines representing a property value (another structure).
        """
        filtered_property_lines = [
            split_line_elements(line)
            for line in text_lines[0][0].split("\n")
            if line != ""
        ]
        return filtered_property_lines

    @classmethod
    def get_tree_structure_read_lines(
        cls, parsed_tuple: Tuple[DSeriesStructure, int]
    ) -> Tuple[DSeriesStructure, int]:
        # Only one line is meant to be read as parsed_text_lines is mapped 1:1 lines-properties.
        return parsed_tuple[0], 1


class DSeriesTreeStructureCollection(DSeriesStructure):
    # TODO Deprecate.
    # This class is overdoing logic that is later on being handled by its parent (DSeriesStructure).
    # It should therefore only be responsible for generating a dictionary of field name - string values,
    # but not object creation.
    @classmethod
    def parse_text(cls, text: str) -> DSeriesStructure:
        """Parses a text into a list of fields that is then
        parsed into the single collection property of this class.
        Example:
            1 - number of structures
                1 - Structure id
                    2 - Number of values in property
                        4 2
        Args:
            text (str): Text to be parsed into collection of structures.

        Raises:
            ValueError: When the number of structures does not match the defined in the text.

        Returns:
            DSeriesStructure: Parsed structure containing collection of other structures.
        """

        parsed_structure, _ = cls.parse_text_lines(
            [split_line_elements(line) for line in text.split("\n") if line != ""]
        )

        return parsed_structure

    @classmethod
    def get_number_of_structures(cls, lines: List[List[str]]) -> int:
        """Provides an overridable method that retrieves the number of
        structures contained in the tree collection.

        Args:
            lines (List[List[str]]): Tree collection lines.

        Returns:
            int: Number of structures expected.
        """
        return int(get_line_property_value(lines[0][0], reversed_key=True))

    @classmethod
    def parse_collection_type(
        cls, collecion_type: Type, lines: list
    ) -> Tuple[DSeriesStructure, int]:
        return collecion_type.parse_text_lines(lines)

    @classmethod
    def parse_text_lines(cls, lines: List[str]) -> Tuple[DSeriesStructure, int]:
        # Get class types
        collection_property_name = list(cls.__fields__.items())[0][0]
        structure_type = get_field_collection_type(cls, 0)
        # Parse elements.
        number_of_structures = cls.get_number_of_structures(lines)
        parsed_structures_collection = []
        read_lines = 1
        for _ in range(number_of_structures):
            if read_lines >= len(lines):
                raise ValueError(
                    f"Expected {number_of_structures} structures,"
                    + " but missing text lines for "
                    + f"{number_of_structures - len(parsed_structures_collection)}."
                )
            parsed_structure, parsed_lines = cls.parse_collection_type(
                structure_type, lines[read_lines:]
            )
            read_lines += parsed_lines
            parsed_structures_collection.append(parsed_structure)

        return cls(**{collection_property_name: parsed_structures_collection}), read_lines


class DSeriesMatrixTreeStructureCollection(DSeriesTreeStructureCollection):
    # TODO Deprecate.
    # This class is overdoing logic that is later on being handled by its parent (DSeriesStructure).
    # It should therefore only be responsible for generating a dictionary of field name - string values,
    # but not object creation.
    @classmethod
    def parse_collection_type(
        cls, collecion_type: Type, lines: list
    ) -> Tuple[DSeriesStructure, int]:
        parsed_structure, _ = collecion_type.parse_text_lines(
            [[line] for line in lines[0]]
        )
        return parsed_structure, 1


# endregion


class DSeriesNoParseSubStructure(DSeriesStructure):
    """DSerie structure that prevents its field from
    being parsed further."""

    @staticmethod
    def is_parseable() -> bool:
        return False


class DSerieParser(BaseParser):
    """Class to parse older DSerie formats in a generic way."""

    def __init__(self, filename: FilePath = None):
        if filename:
            self.datastructure = self.parse(filename)

    @property
    @abstractmethod
    def dserie_structure(self) -> BaseModelStructure:
        raise NotImplementedError("Implement in derived classes.")

    @property
    @abstractmethod
    def suffix_list(self) -> List[str]:
        raise NotImplementedError("Implement in derived classes.")

    def parse(self, filename: FilePath) -> DSeriesStructure:
        logger.debug(f"Parsing {filename}")

        with open(filename) as io:
            datastructure = self.dserie_structure.parse_text(io.read())

        return datastructure

    @staticmethod
    def parse_group_as_dict(text_lines: str) -> Dict[str, str]:
        """Parses a text with headers such as
        [GROUP] and [END OF GROUP] into a dict.

        Arguments:
            text {str} -- List of fields

        Returns:
            dict -- Parsed dictionary.

        Can parse the following:

        [GROUP A]
        data
        more_data
        [END OF GROUP A]

        into:

        {
            "group_a": "data\nmore_data"
        }

        """
        return dict(DSerieParser.parse_group(text_lines=text_lines, unique_keys=True))

    @staticmethod
    def parse_group(
        text_lines: str, loose_properties: bool = False, unique_keys: bool = False
    ) -> Iterable[Tuple[str, str]]:
        """Parses a text with headers such as [GROUP] and [END OF GROUP] and
        yields each property of the group as a Tuple[property name, value].
        Because it is an iterable it may contain repeated elements.
        Delegating responsibility on what to do with such elements to the caller.

        Example:

        [GROUP A]
        data
        more_data
        [END OF GROUP A]

        into:

        (
            "group_a": "data\nmore_data"
        )

        Args:
            text_lines (str): List of fields.
            loose_properties (bool, optional): Properties without a key are yielded.
            unique_keys (bool, optional): Whether elements can be repeated or not.

        Raises:
            Exception: [description]

        Returns:
            Iterable[Tuple[str, str]]: Parsed collection of property values, can include repe.

        Yields:
            Iterator[Iterable[Tuple[str, str]]]: Parsed Tuple[Property name, value]
        """
        currentkey = ""
        data = ""
        for i, line in enumerate(text_lines.split("\n")):
            sline = line.strip()
            # keyline
            if sline.startswith("[") and sline.endswith("]"):
                # [ key name ] => key_name
                key = make_key(sline[1:-1])

                # new group
                if currentkey == "":
                    currentkey = key
                    data = ""

                # duplicate group before end
                elif currentkey == key and unique_keys:
                    raise ValueError(
                        f"Can't parse duplicate key {key} at line {i} without first encountering and END OF."
                    )

                # end of current group
                elif key == "end_of_" + currentkey:
                    yield currentkey, data.strip()
                    data = ""
                    currentkey = ""

                # sub group that is eaten for now
                else:
                    data += line + "\n"

            # dataline
            else:
                if currentkey:
                    data += line + "\n"
                elif not currentkey and loose_properties:
                    yield (currentkey, line)

    @staticmethod
    def parse_list_group(
        text: str,
        exceptions=["vertical", "soil", "residual_settlements"],
        skipped_keys=[],
    ) -> Dict[str, Union[List[str], str]]:
        """Method to parse several groups in a sli/sld file that can include repeated groups.

        Includes an exception to always parse a group as a list.
        Can parse the following:

        [GROUP A]
        data
        [END OF GROUP A]
        [GROUP B]
        [GROUP B1]
        data
        [END OF GROUP B1]
        [END OF GROUP B]
        [GROUP B]
        data
        [END OF GROUP B]

        into

        {
         "group_a": "data\n"
         "group_b: [
            "[GROUP B1]\ndata\n[END OF GROUP B1]",
            "data"
        }

        """

        # Output format
        parsed_dict = {}

        currentkey = ""  # current group we're in
        data = ""  # current collection of lines a group

        text_lines = list(filter(None, text.split("\n")))
        for i, line in enumerate(text_lines):
            sline = line.strip()

            # keyline such as [GROUP]
            if sline.startswith("[") and sline.endswith("]"):
                # [ key name ] => key_name
                key = make_key(sline[1:-1])
                if skipped_keys and any(s_key for s_key in skipped_keys if s_key in key):
                    continue

                # new group
                if currentkey == "":
                    currentkey = key
                    data = ""

                # duplicate group before end
                elif currentkey == key:
                    raise ValueError(
                        f"Can't parse duplicate key at line {i} without first encountering and END OF."
                    )

                # end of current group signaled with [END OF GROUP]
                elif key == "end_of_" + currentkey:

                    # If key already exists, this is a repeated group -> List
                    if currentkey in parsed_dict:
                        # Append in case of a list (2+ elements)
                        if isinstance(parsed_dict[currentkey], list):
                            parsed_dict[currentkey].append(data.strip())
                        # Else create the list first (1 element)
                        else:
                            parsed_dict[currentkey] = [
                                parsed_dict[currentkey],
                                data.strip(),
                            ]
                    # If key doesn't exist yet
                    else:
                        # Check whether the group is always a list
                        if currentkey in exceptions:
                            parsed_dict[currentkey] = [data.strip()]
                        # Otherwise add it in as a single group
                        else:
                            parsed_dict[currentkey] = data.strip()

                    # Reset for next group
                    data = ""
                    currentkey = ""

                # sub group that is eaten for now
                else:
                    data += line + "\n"

            # dataline
            else:
                data += line + "\n"

        return parsed_dict


def make_key(key: str) -> str:
    return (
        key.strip()
        .replace("(", "")
        .replace(")", "")
        .replace("/", "")
        .replace(" - ", "___")
        .replace(" ", "_")
        .replace("-", "__")
        .replace(".", "____")
        .lower()
    )


def strip_line_first_element(text: str):
    return split_line_elements(text)[0]


def split_line_elements(text: str) -> List[str]:
    """Separates by space and tabulator.

    Args:
        text (str): Text to break into separate fields.

    Returns:
        List[str]: List of formatted values.
    """
    # parts = re.split(" |\t", text.strip())
    parts = shlex.split(text.strip())
    values = list(filter(lambda part: part != "", parts))
    return values


def get_line_property_value(text: str, reversed_key: bool) -> str:
    """Returns the property value representation in the dseries files
    assuming is delimited by either ':' or '='.
    Examples:
        Given (reversed_key = True):
            4.2 : property_value
            4.2 = property_value
        Given (reversed_key = False):
            property_value : 4.2
            property_value = 4.2
        Returns:
            4.2

    Args:
        text (str): Unformatted line of text.
        reversed_key (bool): How the key needs to be extracted.

    Returns:
        str: First field containing value.
    """
    _, value = get_line_property_key_value(text, reversed_key)
    return value


def get_line_property_key_value(text: str, reversed_key: bool) -> Tuple[str, str]:
    """Returns the unformatted key and formatted value for a given line.
    Examples:
        reversed_key = False
        key= value 123 -> key, value
        reversed_key = True
        value 123= key asdf -> key, value

    Args:
        text (str): Text from where to extract key and value.
        reversed_key (bool): Whether the key comes before or after the value.

    Returns:
        Tuple[str, str]: Unformatted key and formatted value.
    """
    key_idx, value_idx = (-1, 0) if reversed_key else (0, -1)
    text_parts = re.split(":|=", text.strip())
    key_part = ""
    if len(text_parts) > 1:
        # Only consider a value if there is more than one field.
        key_part = text_parts[key_idx].strip()
    value_part = text_parts[value_idx].strip()
    if value_part:
        # If there was a value in the first field, clean it.
        return key_part, re.split(" ", value_part)[0].strip()
    return key_part, ""


def get_field_collection_type(class_type: Type, field_idx: int) -> Type:
    """Gets the type wrapped by a collection of a given class.
    Example:
        DummyClass:
            dummycollection: List[SubDummyClass]
        get_field_collection_type(DummyClass, 0) -> SubDummyClass

    Args:
        class_type (Type): Class containing a collection field.
        field_idx (int): Position of the collection field in the class.

    Returns:
        Type: The class for the items in the collection.
    """
    list_type = list(class_type.__fields__.items())[field_idx][1]
    if not list_type.sub_fields:
        return list_type.outer_type_
    return list_type.sub_fields[0].outer_type_


def is_structure_collection(field: Type) -> bool:
    return inspect.isclass(field) and issubclass(field, DSeriesStructure)


def read_property_as_list(
    field_name: str, field: Type, text_lines: list
) -> Tuple[List[DSeriesStructure], int]:
    """Reads a property containing a collection of values represented as
    a list of strings.

    Args:
        field_name (str): Name of the field to parse.
        field (Type): Type to parse the values to.
        text_lines (list): List of list of strings containing values.

    Raises:
        ValueError: If the values given don't match the defined expectations.

    Returns:
        Tuple[List[DSeriesStructure], int]: Returns generated DSeriesStructures and lines read.
    """
    # Assumes the values are separated by spaces.
    list_size = int(get_line_property_value(text_lines[0][0], reversed_key=True))
    lines_read = 1
    list_values = []
    # Fetch values from next lines as it might have been split into different lines
    # See GEOLIB-68 to know more about that problem.
    while len(list_values) < list_size:
        if lines_read >= len(text_lines):
            raise ValueError(f"Expected {list_size} values for property {field_name}.")
        iteration_lines = 1
        lines_to_parse = text_lines[lines_read:]
        if issubclass(field, DSeriesTreeStructureCollection):
            # If its an encapsuled collection, number of lines needs to be given.
            lines_to_parse = text_lines[lines_read - 1 :]

        if is_structure_collection(field):
            parsed_values, iteration_lines = field.parse_text_lines(lines_to_parse)
            list_values.append(parsed_values)
        else:
            parsed_values = text_lines[lines_read]
            list_values.extend(parsed_values)
        lines_read += iteration_lines
    return list_values, lines_read



import logging
from os import path, scandir
from typing import List, Tuple, Type, _GenericAlias, get_type_hints
from zipfile import ZipFile

from pydantic import DirectoryPath, FilePath
from zipp import Path

from geolib.models.parsers import BaseParser, BaseParserProvider
from geolib.models.utils import get_filtered_type_hints

from .internal import BaseModelStructure, DStabilityStructure

logger = logging.getLogger(__name__)


class DStabilityParser(BaseParser):
    @property
    def suffix_list(self) -> List[str]:
        return [".json", ""]

    @property
    def structure(self) -> Type[DStabilityStructure]:
        return DStabilityStructure

    def can_parse(self, filename: FilePath) -> bool:
        return super().can_parse(filename) or filename.is_dir()

    def parse(self, filepath: DirectoryPath) -> BaseModelStructure:
        ds = {}

        # Find required .json files via type hints
        for field, fieldtype in get_filtered_type_hints(self.structure):

            # On List types, parse a folder
            if type(fieldtype) == _GenericAlias:  # quite hacky
                element_type, *_ = fieldtype.__args__  # use getargs in 3.8
                ds[field] = self.__parse_folder(element_type, filepath / "")

            # Otherwise it is a single .json in the root folder
            else:
                fn = filepath / (fieldtype.structure_name() + ".json")
                if not fn.exists():
                    raise FileNotFoundError(f"Couldn't find required file at {fn}")
                ds[field] = fieldtype.parse_raw(fn.open().read())

        return self.structure(**ds)

    def __parse_folder(self, fieldtype, filepath: DirectoryPath) -> List:
        out = []
        folder = filepath / fieldtype.structure_group()

        try:
            files = list(folder.iterdir())
        except FileNotFoundError:  # Not all result folders are required.
            return out

        # We need to sort to make sure that files such as x.json, x_1.json,
        # x_2.json etc. are stored sequentally, scandir produces arbitrary order.
        sorted_files = sorted(files, key=lambda x: x.name)
        for file in sorted_files:
            if fieldtype.structure_name() in file.name:
                out.append(fieldtype.parse_raw(file.open().read()))
            else:
                logger.debug(f"Didn't match {fieldtype} for {file}")

        if len(out) == 0:
            logger.debug(f"Couldn't find {fieldtype} file(s) at {folder}")

        return out


class DStabilityZipParser(DStabilityParser):
    @property
    def suffix_list(self) -> List[str]:
        return [".stix"]

    def can_parse(self, filename: FilePath) -> bool:
        return filename.suffix in self.suffix_list

    def parse(self, filepath: FilePath) -> BaseModelStructure:
        with ZipFile(filepath) as zip:

            # Fix backslashes in zipfile (until fixed in DStability)
            for file in zip.filelist:
                new_filename = file.filename.replace("\\", "/")
                if new_filename != file.filename:
                    file.filename = new_filename
            for key in list(zip.NameToInfo.keys()):
                new_key = key.replace("\\", "/")
                if new_key != key:
                    zip.NameToInfo[new_key] = zip.NameToInfo.pop(key)

            path = Path(zip)
            ds = super().parse(path)
        return ds


class DStabilityParserProvider(BaseParserProvider):

    _input_parsers = None
    _output_parsers = None

    @property
    def input_parsers(self) -> Tuple[DStabilityParser, DStabilityZipParser]:
        if not self._input_parsers:
            self._input_parsers = (DStabilityZipParser(), DStabilityParser())
        return self._input_parsers

    @property
    def output_parsers(self) -> Tuple[DStabilityParser, DStabilityZipParser]:
        if not self._output_parsers:
            self._output_parsers = (DStabilityParser(), DStabilityZipParser())
        return self._output_parsers

    @property
    def parser_name(self) -> str:
        return "DStability"



"""
Profile and Layer classes which are used by both D-Foundations and DSheetPiling.
D-Foundations often requires more parameters, which are unused for DSheetPiling.

"""

from typing import List

from pydantic import conlist, validator

from geolib.geometry.one import Point
from geolib.models import BaseDataClass
from geolib.soils import Soil

from .internal import SoilProfile as InternalSoilProfile


class CPT(BaseDataClass):
    """Internal Placeholder CPT measurements as used in D-Foundations/D-Sheetpiling.

    It is here because input files can contain CPTs coupled to a Profile.

    .. todo::
        * Add Friction and other parameters?
        * Unify with DFoundations
    """

    z: List[float]
    qc: List[float]


class CPTRule(BaseDataClass):
    """Placeholder class for CPT interpretation options."""


class SoilLayer(BaseDataClass):
    """Layer in a D-Sheetpiling Profile."""

    top_of_layer: float
    soil: str
    water_pressure_top: float = 0.0
    water_pressure_bottom: float = 0.0


class SoilProfile(BaseDataClass):
    """D-Sheetpiling Profile."""

    name: str
    layers: conlist(SoilLayer, min_items=1)
    coordinate: Point = Point(x=0, y=0)

    @validator("layers")
    def top_of_layers_must_be_decreasing(cls, v):
        top_of_layers = [l.top_of_layer for l in v]
        if top_of_layers != sorted(top_of_layers, reverse=True):
            raise ValueError(
                f"Top of layer must decrease with each layer along depth. Top of layers: {top_of_layers}"
            )
        return v

    def to_internal(self) -> InternalSoilProfile:
        return InternalSoilProfile(**self.dict(exclude_none=True))
